import os

with open('data/long.cov', 'w') as file3:
    with open('data/long_cov.tsv', 'r') as file:
        for idx, line1 in enumerate(file):
            long_values = line1.strip().split("\t")
            del long_values[4::3] # delete extra length values
            long_count = len(long_values[2::2])
            if idx != 0:
                long_cov = sum([float(x) for x in long_values[2::2]])
                tot_depth = long_cov / long_count
                line = (
                    "{contig_name}\t"
                    "{contig_length}\t"
                    "{total_avg_depth}\t"
                    "{long}"
                ).format(
                    contig_name = long_values[0],
                    contig_length = long_values[1],
                    total_avg_depth = tot_depth,
                    long = "\t".join(long_values[2::])
                )
                print(line, file=file3)
            else:
                line = (
                    "{contig_name}\t{contig_length}\ttotalAvgDepth\t{samples}"
                ).format(
                    contig_name = 'contigName',
                    contig_length = 'contigLen',
                    samples = "\t".join(long_values[2::])
                )
                print(line, file=file3)
    # os.remove("data/long_cov.tsv")




import subprocess
import shutil
import os
from pathlib import Path

def checkm(checkm2_db, bin_folder, bin_ext, refinery_max_iterations, output_folder, output_file, threads, log):
    if len([f for f in os.listdir(bin_folder) if f.endswith(bin_ext)]) == 0:
        with open(log, "a") as logf:
            logf.write(f"No bins found in {bin_folder}\n")
        os.makedirs(output_folder)
        Path(output_file).touch()
    elif refinery_max_iterations == 0:
        with open(log, "a") as logf:
            logf.write("Skipping pre-refinery CheckM2 rules\n")
        os.makedirs(output_folder)
        Path(output_file).touch()
    else:
        os.environ["CHECKM2DB"] = f"{checkm2_db}/uniref100.KO.1.dmnd"
        with open(log, "a") as logf:
            logf.write(f"Using CheckM2 database {checkm2_db}/uniref100.KO.1.dmnd\n")
            subprocess.run(
                f"checkm2 predict -i {bin_folder}/ -x {bin_ext} -o {output_folder} -t {threads} --force".split(),
                env=os.environ,
                stdout=logf,
                stderr=subprocess.STDOUT
                )
        shutil.copy(f"{output_folder}/quality_report.tsv", output_file)


if __name__ == '__main__':
    checkm2_db = snakemake.params.checkm2_db_path
    bin_folder = snakemake.params.bin_folder
    bin_ext = snakemake.params.extension
    refinery_max_iterations = snakemake.params.refinery_max_iterations
    output_folder = snakemake.output.output_folder
    output_file = snakemake.output.output_file
    threads = snakemake.threads
    log = snakemake.log[0]

    with open(log, "w") as logf: pass

    checkm(
        checkm2_db,
        bin_folder,
        bin_ext,
        refinery_max_iterations,
        output_folder,
        output_file,
        threads,
        log,
    )



import os
import sys
from subprocess import run, Popen, PIPE, STDOUT
import random
import shutil
import logging
import tempfile

def clean_short_reads(
    cat_or_zcat: str,
    read_path: str,
    read_pair: str,
    output_path: str,
    threads: int,
    log: str,
):
    cat_cmd = f"{cat_or_zcat} {read_path}".split()
    sed_cmd = f"""sed s/@/@{read_pair}_/""".split()

    with open(log, "a") as logf:
        logf.write(f"Shell command: {' '.join(cat_cmd)} | {' '.join(sed_cmd)} > {output_path}\n")
        logf.write(' '.join(sed_cmd))
        with open(output_path, 'a') as out:
            cat = Popen(cat_cmd, stdout=PIPE, stderr=logf)
            sed = Popen(sed_cmd, stdin=cat.stdout, stdout=out, stderr=logf)

            sed.wait()
            cat.wait()

            logf.write(f"cat return: {cat.returncode}\n")
            logf.write(f"sed return: {sed.returncode}\n")

def minimap2_process(
    minimap2_type: str,
    reference: str,
    reads: str,
    threads: int,
    output_paf: str,
    log: str,
):
    minimap2_cmd = f"minimap2 -x {minimap2_type} -t {threads} {reference} {reads}".split()

    with open(log, "a") as logf:
        with open(output_paf, 'w') as out:
            Popen(minimap2_cmd, stdout=out, stderr=logf).wait()

def run_seqkit(
    reads,
    pattern_file: str,
    output_file: str,
    threads: int,
    log: str,
):
    seqkit_cmd = f"seqkit -j {threads} grep  --pattern-file {pattern_file} {reads}".split()
    pigz_cmd = f"pigz -p {threads}".split()

    with open(log, "a") as logf:
        logf.write(f"Shell style: {' '.join(seqkit_cmd)} | {' '.join(pigz_cmd)} > {output_file}\n")

        with open(output_file, 'a') as out:
            seqkit = Popen(seqkit_cmd, stdout=PIPE, stderr=logf)
            pigz = Popen(pigz_cmd, stdin=seqkit.stdout, stdout=out, stderr=logf)
            pigz.wait()
            seqkit.wait()
            logf.write(f"seqkit return: {seqkit.returncode}\n")
            logf.write(f"pigz return: {pigz.returncode}\n")

def run_racon(
    reads: str,
    paf: str,
    reference: str,
    output_file: str,
    threads: int,
    log: str,
):
    racon_cmd = f"racon -m 8 -x -6 -g -8 -w 500 -t {threads} -u {reads} {paf} {reference}".split()

    with open(log, "a") as logf:
        logf.write(' '.join(racon_cmd))
        with open(output_file, 'w') as out:
            Popen(racon_cmd, stdout=out, stderr=logf).wait()


def run_minimap_with_samtools(
    reference: str,
    reads: str,
    threads: int,
    output_file: str,
    log: str,
):

    # write minimap2 output to temporary file

    with open(log, "a") as logf:
        minimap2_cmd = f"minimap2 -ax map-ont -t {threads} {reference} {reads}".split()
        logf.write(' '.join(minimap2_cmd))
        samtools_cmd = f"samtools view -F 4 -b -@ {threads-1} -o {output_file}".split()
        logf.write(' '.join(samtools_cmd))
        minimap2 = Popen(minimap2_cmd, stdout=PIPE, stderr=logf)
        samtools = Popen(samtools_cmd, stdin=minimap2.stdout, stderr=logf)
        samtools.wait()
        minimap2.wait()
        # minimap2.stdout.close()
        # # minimap2.wait()
        # # samtools.stdin.close()
        # samtools.communicate()

    # check if output file exists and is not empty
    with open(log, "a") as logf:
        if os.path.exists(output_file) and os.path.getsize(output_file) > 0:
            logf.write(f"{output_file} created.\n")
            if samtools.returncode == 0:
                logf.write("samtools successfully created bam file.\n")
            else:
                logf.write("samtools failed to create bam file.\n")
                logf.write(f"samtools return: {samtools.returncode}\n")
            return True
        else:
            logf.write(f"Error: {output_file} is empty or does not exist.\n")
            logf.write(f"samtools return: {samtools.returncode}\n")
            return False


def run_polish(
    short_reads_1,
    short_reads_2,
    input_fastq,
    output_dir: str,
    output_prefix: str,
    output_fasta: str,
    polishing_rounds: int,
    medaka_model: str,
    reference: str,
    reference_filter: str,
    max_cov: int,
    illumina: bool,
    long_read_type: str,
    coassemble: bool,
    threads: int,
    log: str,
):
    # out = "data/polishing"

    try:
        os.makedirs(output_dir)
    except FileExistsError:
        pass

    random.seed(89)

    # Whether contigs are polished with illumina or long read
    if illumina:
        if reference_filter != 'none':
            reads = "data/short_reads.fastq.gz"
        elif short_reads_2 != 'none':

            # Racon can't handle paired end reads. It treats them as singled-ended. But when you have paired end reads
            # in separate files they can share the same read name, so we need to alter the read name based on the pair
            if not os.path.exists("data/short_reads.racon.1.fastq.gz"):
                for reads1, reads2 in zip(short_reads_1, short_reads_2):
                    cat_or_zcat1 = "cat"
                    cat_or_zcat2 = "cat"
                    if reads1[-3::] == ".gz":
                        cat_or_zcat1 = "zcat"
                    if reads2[-3::] == ".gz":
                        cat_or_zcat2 = "zcat"

                    clean_short_reads(cat_or_zcat1, reads1, 1, "data/short_reads.racon.1.fastq", threads, log)
                    clean_short_reads(cat_or_zcat2, reads2, 2, "data/short_reads.racon.1.fastq", threads, log)
                    if not coassemble:
                        break

                with open(log, "a") as logf:
                    run(f"pigz -p {threads} --fast data/short_reads.racon.1.fastq".split(), stdout=logf, stderr=STDOUT)

            pe1 = "data/short_reads.racon.1.fastq.gz"
            reads = [pe1]
        else:
            if len(short_reads_1) == 1 or not coassemble:
                pe1 = short_reads_1[0]
            else:
                if not os.path.exists("data/short_reads.1.fastq.gz"):
                    for reads1 in short_reads_1:
                        cat_or_zcat1 = "cat"
                        if reads1[-3::] == ".gz":
                            cat_or_zcat1 = "zcat"
                        
                        with open("data/short_reads.1.fastq", 'a') as out:
                            cat_cmd = f"{cat_or_zcat1} {reads1}".split()
                            with open(log, "a") as logf:
                                Popen(cat_cmd, stdout=out, stderr=logf).wait()

                    with open(log, "a") as logf:
                        run("pigz -p {threads} --fast data/short_reads.1.fastq".split(), stdout=logf, stderr=STDOUT)

                pe1 = "data/short_reads.1.fastq.gz"
            reads = [pe1]
    else:
        reads = input_fastq

    # use racon when using illumina or pacbio data
    if illumina or long_read_type not in ['ont', 'ont_hq']:
        for rounds in range(polishing_rounds):
            paf = os.path.join(output_dir, 'alignment.%s.%d.paf') % (output_prefix, rounds)
            with open(log, "a") as logf:
                logf.write("Generating PAF file: %s for racon round %d...\n" % (paf, rounds))

            # Generate PAF mapping files
            if not os.path.exists(paf): # Check if mapping already exists
                if illumina:
                    if reads != "data/short_reads.fastq.gz":
                        minimap2_process("sr", reference, ' '.join(reads), threads, paf, log)
                    else:
                        minimap2_process("sr", reference, reads, threads, paf, log)
                elif long_read_type in ['ont', 'ont_hq']:
                    sys.exit("ONT reads are not supported for racon polishing")
                else:
                    minimap2_process("map-pb", reference, reads, threads, paf, log)

            cov_dict = {}
            # Populate coverage dictionary,
            with open(paf) as f:
                for line in f:
                    qname, qlen, qstart, qstop, strand, ref, rlen, rstart, rstop = line.split()[:9]
                    qlen, qstart, qstop, rlen, rstart, rstop = map(int, [qlen, qstart, qstop, rlen, rstart, rstop])
                    if ref in cov_dict:
                        cov_dict[ref] += (rstop - rstart) / rlen
                    else:
                        cov_dict[ref] = (rstop - rstart) / rlen

            high_cov = set()
            low_cov = set()
            for i in cov_dict:
                if cov_dict[i] >= max_cov:
                    high_cov.add(i)
                else:
                    low_cov.add(i)

            no_cov = set()
            with open(reference) as ref_file, open(os.path.join(output_dir, "filtered.%s.%d.fa" % (output_prefix, rounds)), 'w') as o:
                for line in ref_file:
                    if line.startswith('>'):
                        name = line.split()[0][1:]
                        if name in low_cov or name in high_cov:
                            o.write(line)
                            getseq = True
                        else:
                            no_cov.add(name)
                            getseq = False
                    elif getseq:
                        o.write(line)

            included_reads = set()
            excluded_reads = set()
            with open(paf) as f, open(os.path.join(output_dir, "filtered.%s.%d.paf" % (output_prefix, rounds)), 'w') as paf_file:
                for line in f:
                    qname, qlen, qstart, qstop, strand, ref, rlen, rstart, rstop = line.split()[:9]
                    qlen, qstart, qstop, rlen, rstart, rstop = map(int, [qlen, qstart, qstop, rlen, rstart, rstop])
                    # if illumina:
                    #     if qname[-2:] in ['/1', '/2']:
                    #         qname = qname[:-2]
                    if ref in low_cov:
                        paf_file.write(line)
                        included_reads.add(qname)
                    elif ref in high_cov:
                        # Down sample reads from high coverage contigs
                        sample_rate = max_cov / cov_dict[ref]
                        if qname in excluded_reads:
                            pass
                        elif qname in included_reads:
                            paf_file.write(line)
                        elif random.random() < sample_rate:
                            included_reads.add(qname)
                            paf_file.write(line)
                        else:
                            excluded_reads.add(qname)
            with open(os.path.join(output_dir, "reads.%s.%d.lst" % (output_prefix, rounds)), "w") as o:
                # if (reads == 'data/short_reads.fastq.gz' or short_reads_2 == 'none') and illumina:
                #     print("Using paired end reads and adding /1 and /2 to read names")
                # else:
                #     print("Using single end reads")

                for i in included_reads:
                    o.write(i + '\n')
                    # if (reads == 'data/short_reads.fastq.gz' or short_reads_2 == 'none') and illumina:
                    #     o.write(i + '/1\n')
                    #     o.write(i + '/2\n')
                    # else:
                    #     o.write(i + '\n')
            logging.info("Retrieving reads...")
            if not isinstance(reads, str):
                for read in reads:
                    pattern_file = f"{output_dir}/reads.{output_prefix}.{rounds}.lst"
                    output_file = f"{output_dir}/reads.{output_prefix}.{rounds}.fastq.gz"
                    run_seqkit(
                        read,
                        pattern_file,
                        output_file=output_file,
                        threads=threads,
                        log=log,
                    )

            else:
                pattern_file = f"{output_dir}/reads.{output_prefix}.{rounds}.lst"
                output_file = f"{output_dir}/reads.{output_prefix}.{rounds}.fastq.gz"
                run_seqkit(
                    reads,
                    pattern_file,
                    output_file=output_file,
                    threads=threads,
                    log=log,
                )

            with open(log, "a") as logf:
                logf.write("Performing round %d of racon polishing..." % rounds)

            reads = f"{output_dir}/reads.{output_prefix}.{rounds}.fastq.gz"
            paf_file = f"{output_dir}/filtered.{output_prefix}.{rounds}.paf"
            reference = f"{output_dir}/filtered.{output_prefix}.{rounds}.fa"
            output_file = f"{output_dir}/filtered.{output_prefix}.{rounds}.pol.fa"

            run_racon(
                reads,
                paf_file,
                reference,
                output_file,
                threads=threads,
                log=log,
            )

            with open(os.path.join(output_dir, "combined.%s.%d.pol.fa" % (output_prefix, rounds)), "w") as o:
                with open(os.path.join(output_dir, "filtered.%s.%d.pol.fa" % (output_prefix, rounds))) as f:
                    gotten_set = set()
                    for line in f:
                        if line.startswith('>'):
                            gotten_set.add(line.split()[0][1:])
                        o.write(line)
                with open(reference) as f:
                    for line in f:
                        if line.startswith('>'):
                            name = line.split()[0][1:]
                            if name in gotten_set:
                                get_line = False
                            else:
                                get_line = True
                        if get_line:
                            o.write(line)
            reference = os.path.join(output_dir, "combined.%s.%d.pol.fa" % (output_prefix, rounds))

        # Copy the final polished reference to the output directory
        shutil.copyfile(reference, output_fasta)
    else:
        # polishing will be done by medaka
        if long_read_type not in ['ont', 'ont_hq']:
            sys.exit("ERROR: long_read_type must be ont or ont_hq for medaka polishing")
        
        # bam = os.path.join(output_dir, 'alignment.%s.1.bam') % (output_prefix)
        # print("Generating BAM file: %s for medaka..." % (bam))
        # # we just run medaka once: https://twitter.com/rrwick/status/1158278701819125760
        # # Twitter is a valid source of information :) do not question.
        # run_minimap_with_samtools(
        #     reference,
        #     reads,
        #     output_file=bam,
        #     threads=threads,
        # )
        with open(log, "a") as logf:
            logf.write("Running medaka...\n")
            medaka_cmd = f"medaka_consensus -t {threads} -i {reads} -m {medaka_model} -o data/polishing/ -d {reference}".split()
            logf.write(' '.join(medaka_cmd))
            run(medaka_cmd, stdout=logf, stderr=STDOUT)

        # copy the output to the expected location

        shutil.copyfile("data/polishing/consensus.fasta", output_fasta)
        # remove the intermediate files
        os.remove("data/polishing/calls_to_draft.bam")
        os.remove("data/polishing/calls_to_draft.bam.bai")
        os.remove("data/polishing/consensus_probs.hdf")


    if os.path.exists("data/short_reads.racon.1.fastq.gz"):
        os.remove("data/short_reads.racon.1.fastq.gz")
        # os.remove("data/short_reads.racon.2.fastq.gz")


if __name__ == "__main__":

    short_reads_1 = snakemake.config["short_reads_1"]
    short_reads_2 = snakemake.config["short_reads_2"]
    input_fastq = snakemake.input.fastq
    reference = snakemake.input.fasta
    reference_filter = snakemake.config["reference_filter"]
    output_dir = "data/polishing"
    output_prefix = snakemake.params.prefix
    output_fasta = snakemake.output.fasta
    rounds = snakemake.params.rounds
    long_read_type = snakemake.config["long_read_type"]
    medaka_model = snakemake.config["medaka_model"]
    illumina = snakemake.params.illumina

    max_cov = snakemake.params.maxcov
    threads = snakemake.threads
    coassemble = snakemake.params.coassemble
    log = snakemake.log[0]

    with open(log, "w") as logf: pass

    run_polish(
        short_reads_1,
        short_reads_2,
        input_fastq,
        reference=reference,
        reference_filter=reference_filter,
        output_dir=output_dir,
        output_prefix=output_prefix,
        output_fasta=output_fasta,
        polishing_rounds=rounds,
        long_read_type=long_read_type,
        medaka_model=medaka_model,
        illumina=illumina,
        max_cov=max_cov,
        threads=threads,
        coassemble=coassemble,
        log=log,
    )


from subprocess import CalledProcessError, run, STDOUT
import os
from pathlib import Path
import logging
import gzip
import subprocess
import tempfile
import glob

class ReadContainer:
    def __init__(self, long_reads, short_reads_1, short_reads_2):
        if long_reads == "none":
            long_reads = []

        if short_reads_1 == "none":
            short_reads_1 = []
        
        if short_reads_2 == "none":
            short_reads_2 = []

        self.long_reads = long_reads
        self.short_reads_1 = short_reads_1
        self.short_reads_2 = short_reads_2
    
    def _check_paired_reads(self):
        if len(self.short_reads_2) == 0:
            return False
        if len(self.short_reads_1) == 0:
            return False
        if len(self.short_reads_1) != len(self.short_reads_2):
            logging.warning("Short read 1 and short read 2 are not the same length but paired reads assumed, please check input")
            return False
        return True

    def get_paired_reads(self):
        if not self._check_paired_reads():
            return []

        paired_reads = []
        for i in range(len(self.short_reads_1)):
            paired_reads.append((self.short_reads_1[i], self.short_reads_2[i]))
        return paired_reads
    
    def get_paired_read_count(self):
        if not self._check_paired_reads() or self._check_interleaved():
            return 0
        return len(self.short_reads_1)

    def get_single_reads(self):
        if self._check_paired_reads() or self._check_interleaved():
            return []
        return self.short_reads_1 if len(self.short_reads_1) > 0 else self.short_reads_2

    def get_single_read_count(self):
        if self._check_paired_reads() or self._check_interleaved():
            return 0
        return len(self.short_reads_1) if len(self.short_reads_1) > 0 else len(self.short_reads_2)

    def get_long_read_count(self):
        return len(self.long_reads)

    def get_long_reads(self):
        return self.long_reads
    
    def get_interleaved_reads(self):
        if not self._check_interleaved() or self._check_paired_reads():
            return []
        return self.short_reads_1 if len(self.short_reads_1) > 0 else self.short_reads_2
    
    def get_interleaved_read_count(self):
        if not self._check_interleaved() or self._check_paired_reads():
            return 0
        return len(self.short_reads_1) if len(self.short_reads_1) > 0 else len(self.short_reads_2)
    
    def _check_interleaved(self):
        if self._check_paired_reads():
            return False
        if len(self.short_reads_1) == 0 and len(self.short_reads_2) == 0:
            return False

        short_reads = self.short_reads_1 if len(self.short_reads_1) > 0 else self.short_reads_2
        # open the reads and check if they are interleaved
        # if they are then return True
        # if they are not then return False
        
        # check if first file is gzipped
        interleaved = False
        if short_reads[0].endswith(".gz"):
            with gzip.open(short_reads[0], "rt") as read_file:
                interleaved = self._forward_and_reverse_present(read_file)
        else:
            with open(short_reads[0], "r") as read_file:
                interleaved = self._forward_and_reverse_present(read_file)

        return interleaved
    
    def _forward_and_reverse_present(self, file) -> bool:
        read_pattern = []
        for i, line in enumerate(file):
            if i > 32:
                break
            if i % 4 == 0:
                if line.endswith("1"):
                    read_pattern.append(1)
                elif line.endswith("2"):
                    read_pattern.append(2)
                else:
                    read_pattern.append(0)
        
        
        for r1, r2 in zip(read_pattern[::2], read_pattern[1::2]):
            if r1 == 0 or r2 == 0:
                return False
            if r1 == r2:
                return False
            if r1 == 1 and r2 == 2:
                return True
        
        return False
    
    def get_total_read_count(self):
        return self.get_long_read_count() + self.get_single_read_count() + self.get_paired_read_count() + self.get_interleaved_read_count()

class SingleMContainer:
    def __init__(self, threads: int, output_dir: str, read_container: ReadContainer, logf: str):
        self.commands = []
        self.threads = threads
        self.read_container = read_container
        self.output_dir = output_dir
        self.intermediate_dir = os.path.join(self.output_dir, "intermediate")
        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(self.intermediate_dir, exist_ok=True)
        self.logf = logf
        self.process_queue = []
        self.total_reads = self.read_container.get_total_read_count()

    
    def run(self):
        with open(self.logf, "a") as logf:
            logf.write("generating SingleM commands\n")
            self.create_commands()
            for command in self.commands:
                logf.write(" ".join(command) + "\n")
            logf.write("running SingleM commands\n")
            self.run_commands(logf)
            self.combine_otu_tables(logf)
    
    def combine_otu_tables(self, logf):
        logf.write("combining SingleM otu tables\n")
        intermediate_otu_tables = glob.glob(os.path.join(self.intermediate_dir, "*.csv"))
        summarise_cmd = f"singlem summarise --input-otu-tables {' '.join(intermediate_otu_tables)} --output-otu-table {os.path.join(self.output_dir, 'metagenome.combined_otu_table.csv')}".split()
        try:
            with open(self.logf, "a") as logf:
                run(summarise_cmd, stdout=logf, stderr=STDOUT)
            Path("data/singlem_out/metagenome.combined_otu_table.csv").touch()
        except CalledProcessError as e:
            with open(log, "a") as logf:
                logf.write(e)
                logf.write("\nSingleM summarise failed. Exiting.\n")
            Path("data/singlem_out/metagenome.combined_otu_table.csv").touch()
   
    def create_commands(self):
        self._create_longread_commands()
        self._create_shortread_commands()


    def run_commands(self, logf):
        process_index = 0
        for command in self.commands:
            f = tempfile.TemporaryFile()
            p = subprocess.Popen(command, stdout=f, stderr=logf)
            self.process_queue.append((p, f))
            process_index += 1
            if len(self.process_queue) >= self.threads:
                self._check_processes(self.threads + 1, logf)
        
        # write how many processes are left
        logf.write(f"waiting for {len(self.process_queue)} processes to finish\n")
        while len(self.process_queue) > 0:
            self._check_processes(0, logf)

    def _check_processes(self, max_processes: int, logf):
        while len(self.process_queue) > max_processes:
            for i, (p, f) in enumerate(self.process_queue):
                if p.poll() is not None:
                    for line in f:
                        logf.write(line)
                    f.close()
                    self.process_queue.pop(i)
                    break


    def _create_longread_commands(self):
        threads = max(self.threads // self.total_reads, 1)
        for i, long_read in enumerate(self.read_container.get_long_reads()):
            command = f"singlem pipe --threads {threads} --sequences {long_read} --otu-table {self.intermediate_dir}/metagenome.longread_{i}_otu_table.csv".split()
            self.commands.append(command)
    
    def _create_shortread_commands(self):
        threads = max(self.threads // self.total_reads, 1)
        for i, short_read in enumerate(self.read_container.get_single_reads()):
            command = f"singlem pipe --threads {threads} --sequences {short_read} --otu-table {self.intermediate_dir}/metagenome.shortread_single_{i}_otu_table.csv".split()
            self.commands.append(command)
        
        for i, (short_read_1, short_read_2) in enumerate(self.read_container.get_paired_reads()):
            command = f"singlem pipe --threads {threads} --forward {short_read_1} --reverse {short_read_2} --otu-table {self.intermediate_dir}/metagenome.shortread_paired_{i}_otu_table.csv".split()
            self.commands.append(command)
        
        for i, interleaved_read in enumerate(self.read_container.get_interleaved_reads()):
            command = f"singlem pipe --threads {threads} --sequences {interleaved_read} --otu-table {self.intermediate_dir}/metagenome.shortread_interleaved_{i}_otu_table.csv".split()
            self.commands.append(command)



def run_singlem(
    read_container: ReadContainer,
    threads: int,
    log: str,
):
    output_dir = "data/singlem_out"
    singlem_container = SingleMContainer(threads, output_dir, read_container, log)
    singlem_container.run()

def valid_path(path: str) -> bool:
    return os.path.exists(path)

if __name__ == '__main__':
    # check if SINGLEM_METAPACKAGE_PATH environment variable is set and path is valid
    # if not then, error and exit
    os.environ["SINGLEM_METAPACKAGE_PATH"] = snakemake.params.package_path
    if "SINGLEM_METAPACKAGE_PATH" not in os.environ or not valid_path(os.environ["SINGLEM_METAPACKAGE_PATH"]):
        raise ValueError("SINGLEM_METAPACKAGE_PATH environment variable not set. Please set using 'aviary configure' or manually. Exiting.")

    long_reads = snakemake.config['long_reads']
    short_reads_1 = snakemake.config['short_reads_1']
    short_reads_2 = snakemake.config['short_reads_2']
    pplacer_threads = snakemake.threads
    log = snakemake.log[0]

    read_container = ReadContainer(long_reads, short_reads_1, short_reads_2)

    with open(log, "w") as logf: pass

    run_singlem(
        read_container,
        pplacer_threads,
        log,
    )



import subprocess
import os
import shutil
import sys
from typing import List



def assemble_short_reads(
        read_set1, read_set2,
        reference_filter: str,
        max_memory: int,
        use_megahit: bool,
        coassemble: bool,
        threads: int,
        tmp_dir: str,
        kmer_sizes: List[int],
        log: str):
    '''
    Assemble short reads using either megahit or spades
    :param read_set1: list of short reads 1, or 'none'
    :param read_set2: list of short reads 2, or 'none'
    :param max_memory: maximum memory to use
    :param use_megahit: use megahit or not
    :param coassemble: coassemble or not
    :param threads: number of threads
    :param tmp_dir: temporary directory
    :param kmer_sizes: list of kmer sizes
    :return:
    '''

    if not tmp_dir:
        try:
            tmp_dir = os.environ["TMPDIR"]
        except KeyError:
            tmp_dir_arg = ""

    if tmp_dir:
        tmp_dir_arg = f"--tmp-dir {tmp_dir}"

    # deal with read sets i.e. are we coassembling? Which assembler are we using?
    # Non co-assembled reads are handled the same for each assembler
    if read_set1 != 'none':
        if not coassemble or len(read_set1) == 1:
            read_set1 = read_set1[0]
            if read_set2 != 'none':
                read_set2 = read_set2[0]

        elif not use_megahit and reference_filter == 'none':
            if read_set2 != 'none':
                for reads1, reads2 in zip(read_set1, read_set2):
                    with open(log, 'a') as logf:
                        with open('data/short_reads.1.fastq.gz', 'a') as out1:
                            subprocess.run(f"cat {reads1}".split(), stdout=out1, stderr=logf)

                        with open('data/short_reads.2.fastq.gz', 'a') as out2:
                            subprocess.run(f"cat {reads2}".split(), stdout=out2, stderr=logf)

                read_set1 = 'data/short_reads.1.fastq.gz'
                read_set2 = 'data/short_reads.2.fastq.gz'
            else:
                for reads1 in read_set1:
                    with open(log, 'a') as logf:
                        with open('data/short_reads.1.fastq.gz', 'a') as out1:
                            subprocess.run(f"cat {reads1}".split(), stdout=out1, stderr=logf)

                read_set1 = 'data/short_reads.1.fastq.gz'

        else:
            read_set1 = ",".join(read_set1)
            if read_set2 != 'none':
                read_set2 = ",".join(read_set2)
    else:
        # forward reads must always be present as either
        # forward or interleaved or single ended
        with open(log, 'a') as logf:
            logf.write(f"============= ERROR =============\n\n")
            logf.write(f"Invalid read sets provided \n\n "
                f"for short read assembly: \n\n")
            logf.write(f"    Set 1: {read_set1}\n\n")
            logf.write(f"    Set 2: {read_set2}\n\n")
            logf.write("Validate read sets and resubmit\n")
        sys.exit(1)


    # designate input read string
    read_string = f"--12 {read_set1}"
    if read_set2 != 'none':
        read_string = f"-1 {read_set1} -2 {read_set2}"
    
    if reference_filter != 'none':
        read_string = f"--12 data/short_reads.fastq.gz"


    # Run chosen assembler
    if use_megahit:
        max_memory_in_bytes = max_memory * 1024*1024*1024
        command = f"megahit {read_string} -t {threads} -m {max_memory_in_bytes} -o data/megahit_assembly {tmp_dir_arg}"

        with open(log, 'a') as logf:
            logf.write(f"Queueing command {command}\n")
            subprocess.run(command.split(), stdout=logf, stderr=subprocess.STDOUT)
        os.makedirs("data/short_read_assembly", exist_ok=True)
        shutil.copyfile("data/megahit_assembly/final.contigs.fa", "data/short_read_assembly/scaffolds.fasta")

    else:
        kmers = " ".join(kmer_sizes)
        command = f"spades.py --memory {max_memory} --meta -t {threads} " \
                f"-o data/short_read_assembly {read_string} -k {kmers} {tmp_dir_arg}"
        with open(log, 'a') as logf:
            logf.write(f"Queueing command {command}\n")
            subprocess.run(command.split(), stdout=logf, stderr=subprocess.STDOUT)


if __name__ == '__main__':
    read_set1 = snakemake.config['short_reads_1']
    read_set2 = snakemake.config['short_reads_2']
    log = snakemake.log[0]

    with open(log, 'w') as logf: pass

    assemble_short_reads(
        read_set1,
        read_set2,
        snakemake.config['reference_filter'],
        snakemake.config['max_memory'],
        snakemake.params.use_megahit,
        snakemake.params.coassemble,
        snakemake.threads,
        snakemake.params.tmpdir,
        snakemake.params.kmer_sizes,
        log,
    )


from subprocess import Popen, PIPE
import os
import multiprocessing as mp
import glob
import logging
from pathlib import Path

def run_mfqe(
    read_pair: str, # either '1' or '2'
    input_reads: str,
    output_file: str,
    log: str,
):
    with open(log, "a") as logf:
        output_prefix = output_file[:-5]
        seqtk_cmd = f"seqtk seq -{read_pair} {input_reads}".split()
        mfqe_cmd = f"mfqe --fastq-read-name-lists {output_file} --output-fastq-files {output_prefix}.{read_pair}.fastq.gz".split()

        seqtk = Popen(seqtk_cmd, stdout=PIPE, stderr=logf)
        mfqe = Popen(mfqe_cmd, stdin=seqtk.stdout, stderr=logf)
        mfqe.wait()

        # thoretically p1 and p2 may still be running, this ensures we are collecting their return codes
        seqtk.wait()
        logf.write(f"seqtk return: {seqtk.returncode}\n")
        logf.write(f"mfqe return: {mfqe.returncode}\n")

def process_short_reads_interleaved_filtered(reads, file, log):
    run_mfqe(read_pair='1', input_reads=reads, output_file=file, log=log)
    run_mfqe(read_pair='2', input_reads=reads, output_file=file, log=log)


def run_seqkit(
    read_pair: str, # either '1' or '2'
    reads: str,
    output_file: str, # also the pattern file for seqkit
    threads: int,
    log: str,
):
    with open(log, "a") as logf:
        output_prefix = output_file[:-5]
        with open(f"{output_prefix}.{read_pair}.fastq.gz", 'w') as out:
            seqtk_cmd = f"seqtk seq -{read_pair} {reads}".split()
            seqkit_cmd = f"seqkit -j {threads} grep -f {output_file}".split()
            pigz_cmd = f"pigz -p {threads}".split()

            seqtk = Popen(seqtk_cmd, stdout=PIPE, stderr=logf)
            seqkit = Popen(seqkit_cmd, stdin=seqtk.stdout, stdout=PIPE, stderr=logf)
            pigz = Popen(pigz_cmd, stdin=seqkit.stdout, stdout=out, stderr=logf)
            pigz.wait()

            seqtk.wait()
            seqkit.wait()
            logf.write(f"seqtk return: {seqtk.returncode}\n")
            logf.write(f"seqkit return: {seqkit.returncode}\n")
            logf.write(f"pigz return: {pigz.returncode}\n")


def process_short_reads_interleaved(reads, file, threads, log):
    run_seqkit(read_pair='1', reads=reads, output_file=file, threads=threads, log=log)
    run_seqkit(read_pair='2', reads=reads, output_file=file, threads=threads, log=log)


def run_seqkit_without_seqtk(
    read_pair: str, # either '1' or '2'
    reads: str,
    output_file: str, # also the pattern file for seqkit
    threads: int,
    log: str,
):
    with open(log, "a") as logf:
        output_prefix = output_file[:-5]
        with open(f"{output_prefix}.{read_pair}.fastq.gz", 'w') as out:
            seqkit_cmd = f"seqkit -j {threads} grep -f {output_file} {reads}".split()
            pigz_cmd = f"pigz -p {threads}".split()

            seqkit = Popen(seqkit_cmd, stdout=PIPE, stderr=logf)
            pigz = Popen(pigz_cmd, stdin=seqkit.stdout, stdout=out, stderr=logf)
            pigz.wait()

            seqkit.wait()
            logf.write(f"seqkit return: {seqkit.returncode}\n")
            logf.write(f"pigz return: {pigz.returncode}\n")

def process_short_reads_paired(reads_1, reads_2, file, threads, log):
    run_seqkit_without_seqtk(read_pair='1', reads=reads_1, output_file=file, threads=threads, log=log)
    run_seqkit_without_seqtk(read_pair='2', reads=reads_2, output_file=file, threads=threads, log=log)



def process_long_reads(reads, file, threads, log):
    with open(log, "a") as logf:
        output_prefix = file[:-5]
        with open(f"{output_prefix}.fastq.gz", 'w') as out:
            seqtk_cmd = f"seqtk subseq {reads} {file}".split()
            pigz_cmd = f"pigz -p {threads}".split()

            seqtk = Popen(seqtk_cmd, stdout=PIPE, stderr=logf)
            pigz = Popen(pigz_cmd, stdin=seqtk.stdout, stdout=out, stderr=logf)
            pigz.wait()

            seqtk.wait()
            logf.write(f"seqtk return: {seqtk.returncode}\n")
            logf.write(f"pigz return: {pigz.returncode}\n")


def get_index(n_files, current):
    if current >= n_files / 2:
        return 2
    else:
        return 1

def get_binned_reads(
    long_reads,
    short_reads_1,
    short_reads_2,
    threads: int,
    output_file: str,
    log: str,
):
    if long_reads != 'none':
        logging.info("Extracting binned long reads...")

        pool = mp.Pool(threads)

        mp_results = [pool.apply_async(process_long_reads,
                                        args=(long_reads[0], file, 2, log))
                                        for file in glob.glob('data/binned_reads/*.long.list')]
        for result in mp_results:
            result.get()

        pool.close()
        pool.join()

    if os.path.exists('data/short_reads.fastq.gz'):
        logging.info("Extracting binned interleaved filtered short reads...")
        pool = mp.Pool(threads)
        # extract reads from filtered short reads
        # take end 1 and 2 separately
        mp_results = [pool.apply_async(process_short_reads_interleaved_filtered,
                                        args=('data/short_reads.fastq.gz', file, log))
                                        for file in glob.glob('data/binned_reads/*.short.list')]
        for result in mp_results:
            result.get()

        pool.close()
        pool.join()

    elif short_reads_2 != 'none':
        logging.info("Extracting binned paired end short reads...")
        pool = mp.Pool(threads // 2)
        # extract reads from all short read files
        mp_results = [pool.apply_async(process_short_reads_paired,
                                        args=(' '.join(short_reads_1),
                                            ' '.join(short_reads_2),
                                            file, 2, log))
                                        for file in glob.glob('data/binned_reads/*.short.list')]

        for result in mp_results:
            result.get()

        pool.close()
        pool.join()

    elif short_reads_1 != 'none':
        logging.info("Extracting binned interleaved short reads...")
        pool = mp.Pool(threads // 2)
        # extract reads from all short read files
        mp_results = [pool.apply_async(process_short_reads_interleaved,
                                    args=(' '.join(short_reads_1),
                                            file, 2, log))
                    for file in glob.glob('data/binned_reads/*.short.list')]

        for result in mp_results:
            result.get()

        pool.close()
        pool.join()

    Path(output_file).touch()


if __name__ == '__main__':
    log = snakemake.log[0]
    with open(log, "w") as logf: pass

    get_binned_reads(
        snakemake.config['long_reads'],
        snakemake.config['short_reads_1'],
        snakemake.config['short_reads_2'],
        snakemake.threads,
        snakemake.output,
        log=log
    )



#!/usr/bin/env python3

import unittest
import os
import tempfile
from aviary.modules.binning.scripts.run_checkm import checkm
from unittest.mock import patch, ANY
import subprocess
from pathlib import Path

def create_output(_, env, stdout, stderr):
    os.makedirs("output_folder")
    Path(os.path.join("output_folder", "quality_report.tsv")).touch()

class Tests(unittest.TestCase):
    def test_run_checkm(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            with patch.object(subprocess, "run", side_effect=create_output) as mock_subprocess:
                checkm2_db = os.path.join("checkm2_db")
                os.makedirs(checkm2_db)
                bin_folder = os.path.join("bin_folder")
                os.makedirs(bin_folder)
                Path(os.path.join(bin_folder, "bin_1.fna")).touch()

                checkm(checkm2_db, bin_folder, "fna", 1, "output_folder", "output_file", 1, "checkm.log")
                self.assertTrue(os.path.exists("output_file"))
                os.environ["CHECKM2DB"] = f"{checkm2_db}/uniref100.KO.1.dmnd"

                mock_subprocess.assert_called_once_with(
                    f"checkm2 predict -i {bin_folder}/ -x fna -o output_folder -t 1 --force".split(),
                    env=os.environ,
                    stdout=ANY,
                    stderr=subprocess.STDOUT
                    )

    def test_run_checkm_no_bins(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            with patch.object(subprocess, "run") as mock_subprocess:
                checkm2_db = os.path.join("checkm2_db")
                os.makedirs(checkm2_db)
                bin_folder = os.path.join("bin_folder")
                os.makedirs(bin_folder)

                checkm(checkm2_db, bin_folder, "fna", 1, "output_folder", "output_file", 1, "checkm.log")
                self.assertTrue(os.path.exists("output_file"))
                mock_subprocess.assert_not_called()

    def test_run_checkm_no_refinery(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.chdir(tmpdir)
            with patch.object(subprocess, "run") as mock_subprocess:
                checkm2_db = os.path.join("checkm2_db")
                os.makedirs(checkm2_db)
                bin_folder = os.path.join("bin_folder")
                os.makedirs(bin_folder)
                Path(os.path.join(bin_folder, "bin_1.fna")).touch()

                checkm(checkm2_db, bin_folder, "fna", 0, "output_folder", "output_file", 1, "checkm.log")
                self.assertTrue(os.path.exists("output_file"))
                mock_subprocess.assert_not_called()


if __name__ == '__main__':
    unittest.main()



import os
from pathlib import Path
import glob

def check_and_remove_base_file(file_path) -> str:
    file_name = os.path.basename(file_path)
    if os.path.exists(file_name):
        os.remove(file_name)

    return file_name


if __name__ == '__main__':
    final_bins = snakemake.input.final_bins
    coverage_file = snakemake.input.coverm
    contig_coverage = snakemake.input.contig_coverage
    gtdbtk = snakemake.input.gtdbtk
    singlem = snakemake.input.singlem

    output_taxonomy = "taxonomy"
    output_singlem = "diversity"

    os.chdir('bins/')

    if len(coverage_file) > 0:
        file_name = check_and_remove_base_file(coverage_file)
        os.symlink(f"../{coverage_file}", f"{file_name}")
    
    if len(contig_coverage) > 0:
        file_name = check_and_remove_base_file(contig_coverage)
        os.symlink(f"../{contig_coverage}", f"{file_name}")
    
    os.chdir('..')
    if len(gtdbtk) > 0:
        check_and_remove_base_file(output_taxonomy)
        os.symlink(f"{os.path.dirname(gtdbtk)}", output_taxonomy)
    if len(singlem) > 0:
        check_and_remove_base_file(output_singlem)
        os.symlink(f"{os.path.dirname(singlem)}", output_singlem)
    
    for f in glob.glob('data/binning_bams/*.ba*'):
        os.remove(f)

    Path('bins/done').touch()



import subprocess
import os
import sys
from typing import List

def spades_asssembly(
    input_fastq: str,
    input_long_reads: str,
    output_fasta: str,
    output_spades_folder: str,
    max_memory: int,
    threads: int,
    kmer_sizes: List[int],
    tmp_dir: str,
    long_read_type: str,
    log: str):

    '''
    Assemble short reads and long reads (if any) using spades
    :param input_fastq: short reads fastq file
    :param input_long_reads: long reads fastq file
    :param output_fasta: output fasta file
    :param output_spades_folder: output spades folder
    :param max_memory: maximum memory to use
    :param threads: number of threads
    :param tmpdir: temporary directory
    :param kmer_sizes: list of kmer sizes
    :param long_read_type: type of long reads
    :param log: log file
    :return:
    '''
    if tmp_dir:
        tmp_dir_arg = f"--tmp-dir {tmp_dir}"
    else:
        try:
            tmp_dir = os.environ["TMPDIR"]
            tmp_dir_arg = f"--tmp-dir {tmp_dir}"
        except KeyError:
            tmp_dir_arg = ""
    
    if os.path.exists("data/spades_assembly/tmp"):
        with open(log, 'a') as logf:
            subprocess.run("rm -rf data/spades_assembly/tmp".split(), stdout=logf, stderr=subprocess.STDOUT)   
    # remove existing temporary directory
    minimumsize=500000
    actualsize = int(subprocess.check_output('stat -c%s data/short_reads.filt.fastq.gz', shell=True))
    # check if directory exists
    if  os.path.exists("data/spades_assembly"):
        # resume previous assembly
        command = f"spades.py --restart-from last --memory {max_memory} -t {threads} " \
                f"-o data/spades_assembly -k {kmer_sizes}  {tmp_dir_arg}" 
        # run cmd
        with open(log, 'a') as logf:
            logf.write(f"Queueing command {command}\n")
            subprocess.run(command.split(), stdout=logf, stderr=subprocess.STDOUT)
            subprocess.run("cp data/spades_assembly/scaffolds.fasta data/spades_assembly.fasta".split(), stdout=logf, stderr=subprocess.STDOUT)
    elif actualsize >= minimumsize:
        if long_read_type in ["ont","ont_hq"]:
            command = f"spades.py --checkpoints all --memory {max_memory} --meta --nanopore {input_long_reads} --12 {input_fastq} "\
                f"-o data/spades_assembly -t {threads}  -k {kmer_sizes} {tmp_dir_arg} "       
        else:
            command = f"spades.py --checkpoints all --memory {max_memory} --meta --pacbio {input_long_reads} --12 {input_fastq} "\
                f"-o data/spades_assembly -t {threads}  -k {kmer_sizes} {tmp_dir_arg} "
        # run cmd
        with open(log, 'a') as logf:
            logf.write(f"Queueing command {command}\n")
            subprocess.run(command.split(), stdout=logf, stderr=subprocess.STDOUT)        
            subprocess.run("cp data/spades_assembly/scaffolds.fasta data/spades_assembly.fasta".split(), stdout=logf, stderr=subprocess.STDOUT)
    else:
        with open(log, 'a') as logf:
            subprocess.run(f"mkdir -p {output.spades_folder} && touch {output.fasta}".split(), stdout=logf, stderr=subprocess.STDOUT)
    

if __name__ == '__main__':
    log = snakemake.log[0]
    with open(log, 'w') as logf: pass

    spades_asssembly(
        snakemake.input.fastq,
        snakemake.input.long_reads,
        snakemake.output.fasta,
        snakemake.output.spades_folder,
        snakemake.params.max_memory,
        snakemake.threads,
        snakemake.params.kmer_sizes,
        snakemake.params.tmpdir,
        snakemake.params.long_read_type,
        log
    )


from subprocess import CalledProcessError, run, STDOUT
import os
from pathlib import Path
import logging
import gzip
import subprocess
import tempfile
import glob


class SingleMContainer:
    def __init__(self, threads: int, output_dir: str, genomes: str, assembly: str, pipe_results: str, logf: str):
        self.commands = []
        self.threads = threads
        self.pipe_results = pipe_results
        self.genomes = []
        # for each file ending in .fna in genomes folder, add to self.genomes
        for file in os.listdir(genomes):
            if file.endswith(".fna"):
                self.genomes.append(f"{genomes}/{file}")
        self.assembly = assembly
        self.output_dir = output_dir
        self.intermediate_dir = os.path.join(self.output_dir, "intermediate_genomes")
        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(self.intermediate_dir, exist_ok=True)
        self.logf = logf
        self.process_queue = []

    
    def run(self):
        with open(self.logf, "a") as logf:
            logf.write("generating SingleM commands\n")
            self.create_commands()
            for command in self.commands:
                logf.write(" ".join(command) + "\n")
            logf.write("running SingleM commands\n")
            self.run_commands(logf)
            self.appraise_otu_tables(logf)
    
    def appraise_otu_tables(self, logf):
        logf.write("Appraising SingleM otu tables\n")
        genome_otu_tables = glob.glob(os.path.join(self.intermediate_dir, "*genome_single*.csv"))
        assembly_otu_table = os.path.join(self.intermediate_dir, "metagenome.assembly_0_otu_table.csv")
        appraise_cmd = f"singlem appraise --genome-otu-tables {' '.join(genome_otu_tables)} \
            --assembly-otu-table {assembly_otu_table} \
            --metagenome-otu-tables {self.pipe_results} \
            --plot {os.path.join(self.output_dir, 'singlem_appraise.svg')} \
            --output-binned-otu-table {os.path.join(self.output_dir, 'binned.otu_table.csv')} \
            --output-unbinned-otu-table {os.path.join(self.output_dir, 'unbinned.otu_table.csv')} \
            --output-assembled-otu-table {os.path.join(self.output_dir, 'assembled.otu_table.csv')}".split()
        try:
            # create output file: data/singlem_out/singlem_appraisal.tsv
            output_file = os.path.join(self.output_dir, "singlem_appraisal.tsv")
            with open(output_file, "w") as outf:
                with open(self.logf, "a") as logf:
                    run(appraise_cmd, stdout=outf, stderr=logf)
            Path("data/singlem_out/metagenome.combined_otu_table.csv").touch()
        except CalledProcessError as e:
            with open(log, "a") as logf:
                logf.write(e)
                logf.write("\nSingleM appraise failed. Exiting.\n")
            Path("data/singlem_out/metagenome.combined_otu_table.csv").touch()
   
    def create_commands(self):
        self._create_genome_commands()
        self._create_assembly_commands()


    def run_commands(self, logf):
        process_index = 0
        for command in self.commands:
            f = tempfile.TemporaryFile()
            logf.write(f"running process {process_index} {' '.join(command)}\n")
            p = subprocess.Popen(command, stdout=f, stderr=logf)
            self.process_queue.append((p, f))
            process_index += 1
            if len(self.process_queue) >= self.threads:
                self._check_processes(self.threads + 1, logf)
        
        # write how many processes are left
        logf.write(f"waiting for {len(self.process_queue)} processes to finish\n")
        while len(self.process_queue) > 0:
            self._check_processes(0, logf)

    def _check_processes(self, max_processes: int, logf):
        while len(self.process_queue) > max_processes:
            for i, (p, f) in enumerate(self.process_queue):
                if p.poll() is not None:
                    for line in f:
                        logf.write(line)
                    f.close()
                    self.process_queue.pop(i)
                    break


    def _create_assembly_commands(self):
        threads = max(self.threads // (len(self.genomes) + 1), 1)
        command = f"singlem pipe --threads {threads} --genome-fasta-files {self.assembly} --otu-table {self.intermediate_dir}/metagenome.assembly_0_otu_table.csv".split()
        self.commands.append(command)
    
    def _create_genome_commands(self):
        threads = max(self.threads // (len(self.genomes) + 1), 1)
        for i, genome in enumerate(self.genomes):
            command = f"singlem pipe --threads {threads} --genome-fasta-files {genome} --otu-table {self.intermediate_dir}/metagenome.genome_single_{i}_otu_table.csv".split()
            self.commands.append(command)


def run_singlem(
    genomes_folder: str,
    assembly: str,
    pipe_results: str,
    threads: int,
    log: str,
):
    output_dir = "data/singlem_out"
    singlem_container = SingleMContainer(threads, output_dir, genomes_folder, assembly, pipe_results, log)
    singlem_container.run()

def valid_path(path: str) -> bool:
    return os.path.exists(path)

if __name__ == '__main__':
    # check if SINGLEM_METAPACKAGE_PATH environment variable is set and path is valid
    # if not then, error and exit
    os.environ["SINGLEM_METAPACKAGE_PATH"] = snakemake.params.package_path
    if "SINGLEM_METAPACKAGE_PATH" not in os.environ or not valid_path(os.environ["SINGLEM_METAPACKAGE_PATH"]):
        raise ValueError("SINGLEM_METAPACKAGE_PATH environment variable not set. Please set using 'aviary configure' or manually. Exiting.")

    assembly = snakemake.input.assembly
    genomes = snakemake.params.genomes_folder
    pipe_results = snakemake.input.pipe_results
    threads = snakemake.threads
    log = snakemake.log[0]

    with open(log, "w") as logf: pass

    run_singlem(
        genomes,
        assembly,
        pipe_results,
        threads,
        log,
    )



import os
import subprocess

def assemble_pools(
    input_list: str,
    input_fasta: str,
    output_fasta: str,
    metabat_done: str,
    threads: int,
    log: str):
    """
    Assemble metagenome bins using unicycler
    """
        
    try:
        os.makedirs("data/final_assemblies")
    except FileExistsError:
        pass

    out_assemblies = []
    with open(input_list) as f:
        for line in f:
            if len(line.split()) == 6:
                mb_bin, long_reads, length, bases_nano, short_reads, bases_ill = line.split()
            else:
                mb_bin, long_reads, length, bases_nano = line.split()
                short_reads, bases_ill = 'none', 0
            if os.stat(long_reads).st_size == 0:
                no_long = True
            else:
                no_long = False
            long_reads = long_reads[:-5] + '.fastq.gz'
            long_reads = os.path.abspath(long_reads)
            short_reads_1 = short_reads[:-5] + '.1.fastq.gz'
            short_reads_2 = short_reads[:-5] + '.2.fastq.gz'
            short_reads_1 = os.path.abspath(short_reads_1)
            short_reads_2 = os.path.abspath(short_reads_2)
            length, bases_nano, bases_ill = float(length), float(bases_nano), float(bases_ill)
            out_assemblies.append('data/final_assemblies/%s_unicyc/assembly.fasta' % mb_bin)
            if not os.path.exists('data/final_assemblies/%s_unicyc/assembly.fasta' % mb_bin):
                with open(log, 'a') as logf:
                    if short_reads == 'none':
                        subprocess.run(
                            f"unicycler --verbosity 0 -t {threads} -l {long_reads} -o data/final_assemblies/{mb_bin}_unicyc".split(),
                            stdout=logf, stderr=subprocess.STDOUT
                            )
                    elif no_long:
                        subprocess.run(
                            f"unicycler --verbosity 0 -t {threads} -1 {short_reads_1} -2 {short_reads_2} -o data/final_assemblies/{mb_bin}_unicyc".split(),
                            stdout=logf, stderr=subprocess.STDOUT
                            )
                    else:
                        subprocess.run(
                            f"unicycler --verbosity 0 -t {threads} -1 {short_reads_1} -2 {short_reads_2} -l {long_reads} -o data/final_assemblies/{mb_bin}_unicyc".split(),
                            stdout=logf, stderr=subprocess.STDOUT
                            )

    unbinned_set = set()
    if os.path.exists(metabat_done[:-4] + "binned_contigs.unbinned"):
        with open(metabat_done[:-4] + "binned_contigs.unbinned") as f:
            for line in f:
                unbinned_set.add(line.rstrip())


    with open(output_fasta, 'w') as o:
        count = 0
        getseq = False
        with open(input_fasta) as f:
            for line in f:
                if line.startswith('>') and line.split()[0][1:] in unbinned_set:
                    getseq = True
                    o.write('>unbinned_' + str(count) + '\n')
                    count += 1
                elif line.startswith('>'):
                    getseq = False
                elif getseq:
                    o.write(line)
        for i in out_assemblies:
            if not os.path.exists(i):
                with open(i[:-14] + 'unicycler.log') as f:
                    lastline = f.readlines()[-1]
                    if lastline.startswith("Error: SPAdes failed to produce assemblies. See spades_assembly/assembly/spades.log for more info.") or \
                        lastline.startswith("Error: none of the SPAdes graphs were suitable for scaffolding in Unicycler") or \
                        lastline.startswith("Error: miniasm assembly failed"):
                        continue
            with open(i) as assembly:
                for line in assembly:
                    if line.startswith('>'):
                        count += 1
                        o.write('>unicycler_' + str(count) + '\n')
                    else:
                        o.write(line)

    if not os.path.exists(output_fasta):
        open(output_fasta, 'a').close()



if __name__ == '__main__':
    input_list = snakemake.input.list
    input_fasta = snakemake.input.fasta
    output_fasta = snakemake.output.fasta
    metabat_done = snakemake.input.metabat_done

    threads = snakemake.threads
    log = snakemake.log[0]

    with open(log, 'w') as logf: pass

    assemble_pools(input_list, input_fasta, output_fasta, metabat_done, threads, log)


import logging
import os
import sys

import extern

if __name__ == '__main__':
    unrefined_binners_to_use = [
        ('concoct', 'fa'),
        ('maxbin2', 'fasta'),
        ('vamb', 'fna')]
    refined_binners_to_use = [
        ('rosella', 'fna'),
        ('semibin', 'fna')]

    # N.B. specifying "metabat" will skip both MetaBAT1 and MetaBAT2.
    metabats = ['metabat_sspec', 'metabat_ssens', 'metabat_sens', 'metabat_spec']

    binners = []
    for (binner, extension) in unrefined_binners_to_use:
        if binner not in snakemake.config['skip_binners']:
            binners.append((f'{binner}_bins/', extension, f'data/{binner}_bins.tsv'))

    for (binner, extension) in refined_binners_to_use:
        if binner not in snakemake.config['skip_binners']:
            binners.append((binner+'_refined/final_bins/', extension, f'data/{binner}_refined_bins.tsv'))

    for metabat in metabats:
        if metabat not in snakemake.config['skip_binners']:
            binners.append((metabat.replace('metabat','metabat_bins'), 'fa', f'data/{metabat}_bins.tsv'))
    if 'metabat2' not in snakemake.config['skip_binners']:
        binners.append(('metabat2_refined/final_bins/', 'fna', f'data/metabat2_refined_bins.tsv'))

    logfile = snakemake.log[0]
    logging.basicConfig(filename=logfile, level=logging.INFO)
    logging.info("Using the following binners: " + str(binners))

    if len(binners) == 0:
        logging.error("All binners have been skipped, so DAS_tool cannot be run.")
        sys.exit(1)

    bin_definition_files = []
    for binner, extension, bin_definition_file in binners:
        extern.run(f'Fasta_to_Scaffolds2Bin.sh -i data/{binner} -e {extension} >{bin_definition_file}  2>> {logfile}')
        if os.path.getsize(bin_definition_file) == 0:
            logging.warning(f'Bin definition file {bin_definition_file} is empty, suggesting that {binner} failed or did not not create any output bins.')
        else:
            bin_definition_files.append(bin_definition_file)
    logging.info("Bin definition files created: " + str(bin_definition_files))

    scaffold2bin_files = ','.join(bin_definition_files)

    das_tool_command = f'DAS_Tool --search_engine diamond --write_bin_evals 1 --write_bins 1 -t {snakemake.threads} --score_threshold -42 \
        -i {scaffold2bin_files} \
        -c {snakemake.input.fasta} \
        -o data/das_tool_bins_pre_refine/das_tool >> {logfile} 2>&1'
    logging.info("Running DAS_Tool with command: " + das_tool_command)
    extern.run(das_tool_command)



import os
from subprocess import Popen, PIPE
from pathlib import Path
from typing import List

def run_skip_qc(
    long_reads: List[str],
    output_long_reads: str,
    coassemble: bool,
    log_file: str
):
    """
    Skip quality control
    :param long_reads: list of long reads
    :param output_long_reads: output long reads
    :param coassemble: coassemble or not, if true we will filter all reads into the same file
    :return:
    """
    with open(log_file, 'a') as logf:
        logf.write(f"Skipping quality control\n")
        logf.write(f"Coassemble: {coassemble}\n")
        if coassemble:
            # cat files into the same file
            with open(output_long_reads, 'w') as out:
                for reads in long_reads:
                    if not os.path.exists(reads):
                        logf.write(f"Long read file {reads} does not exist\n")
                        exit(1)

                    cat_or_zcat = 'zcat' if reads.endswith('.gz') else 'cat'
                    cat_cmd = f'{cat_or_zcat} {reads}'.split()
                    pigz_cmd = f'pigz -p {threads}'.split()

                    logf.write(f"Shell style : {' '.join(cat_cmd)} | {' '.join(pigz_cmd)} > {output_long_reads}\n")

                    cat_p1 = Popen(cat_cmd, stdout=PIPE, stderr=logf)
                    pigz_p2 = Popen(pigz_cmd, stdin=cat_p1.stdout, stdout=out, stderr=logf)

                    pigz_p2.wait()
                    cat_p1.wait()
                    logf.write(f"cat return: {cat_p1.returncode}\n")
                    logf.write(f"pigz return: {pigz_p2.returncode}\n")


        elif len(long_reads) > 0:
            # sym link the first file as the output file
            logf.write(f"Symlinking {long_reads[0]} to {output_long_reads}\n")
            if not os.path.exists(long_reads[0]):
                logf.write(f"Long read file {long_reads[0]} does not exist\n")
                exit(1)

            os.symlink(long_reads[0], output_long_reads)
        else:
            logf.write(f"No long reads to quality control\n")
            Path(output_long_reads).touch()

def qc_long_reads(
    long_reads: List[str], # long reads to quality control
    reference_filter: List[str], # one or more references to filter against
    coassemble: bool,
    skip_qc: bool,
    min_length: int,
    min_quality: int,
    keep_percent: int,
    threads: int,
    output_long_reads: str,
    log_file: str,
):
    """
    Quality control long reads using chopper
    :param long_reads: list of long reads
    :param reference_filter: list of reference genomes to filter against
    :param coassemble: coassemble or not, if true we will filter all reads into the same file
    :param min_length: minimum length of reads to keep
    :param min_quality: minimum mean quality of reads to keep
    :param keep_percent: percent of reads to keep
    :param threads: number of threads
    :param output_long_reads: output long reads
    :return:
    """

    if skip_qc or len(long_reads) == 0:
        run_skip_qc(long_reads, output_long_reads, coassemble, log_file)
        return

    # if we have more than one reference_filter file, we need to concatenate them into a single temp file
    with open(log_file, 'a') as logf:
        reference_filter_file_string = ''
        if len(reference_filter) > 1:
            with open(f'{output_long_reads}.reference_filter.fasta', 'w') as out:
                for reference in reference_filter:
                    # check if file exists
                    if not os.path.exists(reference):
                        logf.write(f"Reference filter file {reference} does not exist\n")
                        exit(1)

                    # concatenate accoutning for gzipped files
                    cat_or_zcat = 'zcat' if reference.endswith('.gz') else 'cat'
                    cat_cmd = f'{cat_or_zcat} {reference}'.split()

                    logf.write(f"Shell style : {' '.join(cat_cmd)} > {output_long_reads}\n")

                    cat_p1 = Popen(cat_cmd, stdout=out, stderr=logf)
                    cat_p1.wait()
                    logf.write(f"cat return: {cat_p1.returncode}\n")

            # gzip the concatenated file
            pigz_cmd = f'pigz -p {threads} {output_long_reads}.reference_filter.fasta'.split()
            logf.write(f"Shell style : {' '.join(pigz_cmd)}\n")

            pigz_p1 = Popen(pigz_cmd, stderr=logf)
            pigz_p1.wait()
            logf.write(f"pigz return: {pigz_p1.returncode}\n")

            reference_filter_file_string = f'-c {output_long_reads}.reference_filter.fasta.gz'
        elif len(reference_filter) == 1:
            # make sure file exists
            if not os.path.exists(reference_filter[0]):
                logf.write(f"Reference filter file {reference_filter[0]} does not exist\n")
                exit(1)

            reference_filter_file_string = f'-c {reference_filter[0]}'


        # run chopper    
        # chopper reads on stdin and write to stdout
        logf.write(f"Running chopper on {len(long_reads)} files\n")
        logf.write(f"Reference filter: {reference_filter_file_string}\n")
        with open(output_long_reads, 'w') as out:
            for long_read in long_reads:
                # make sure file exists
                if not os.path.exists(long_read):
                    logf.write(f"Long read file {long_read} does not exist\n")
                    exit(1)

                cat_or_zcat = 'zcat' if long_read.endswith('.gz') else 'cat'
                cat_cmd = f'{cat_or_zcat} {long_read}'.split()
                chopper_cmd = f'chopper -l {min_length} -q {min_quality} -t {threads} {reference_filter_file_string}'.split()
                pigz_cmd = f'pigz -p {threads}'.split()

                logf.write(f"Shell style : {' '.join(cat_cmd)} | {' '.join(chopper_cmd)} | {' '.join(pigz_cmd)} > {output_long_reads}\n")
                
                cat_p1 = Popen(cat_cmd, stdout=PIPE, stderr=logf)
                chopper_p2 = Popen(chopper_cmd, stdin=cat_p1.stdout, stdout=PIPE, stderr=logf)
                pigz_p3 = Popen(pigz_cmd, stdin=chopper_p2.stdout, stdout=out, stderr=logf)

                pigz_p3.wait()
                chopper_p2.wait()
                cat_p1.wait()
                logf.write(f"cat return: {cat_p1.returncode}\n")
                logf.write(f"chopper return: {chopper_p2.returncode}\n")
                logf.write(f"pigz return: {pigz_p3.returncode}\n")

                if not coassemble:
                    break
        
        # clean up reference filter if we concatenated
        if os.path.exists(f'{output_long_reads}.reference_filter.fasta.gz'):
            os.remove(f'{output_long_reads}.reference_filter.fasta.gz')


if __name__ == '__main__':
    # inputs
    long_reads = snakemake.input.long_reads

    # params
    coassemble = snakemake.params.coassemble
    min_length = snakemake.params.min_length
    min_quality = snakemake.params.min_mean_q
    keep_percent = snakemake.params.keep_percent
    reference_filter = snakemake.params.reference_filter
    skip_qc = snakemake.params.skip_qc
    threads = snakemake.threads

    # output
    output_long = snakemake.output.output_long

    # log
    log = snakemake.log[0]
    with open(log, "w") as logf: pass

    qc_long_reads(
        long_reads=long_reads,
        reference_filter=reference_filter,
        coassemble=coassemble,
        skip_qc=skip_qc,
        min_length=min_length,
        min_quality=min_quality,
        keep_percent=keep_percent,
        threads=threads,
        output_long_reads=output_long,
        log_file=log,
    )


from subprocess import run, Popen, STDOUT
import multiprocessing as mp
import os
from pathlib import Path


def spawn_fastqc(file, log, threads=1):
    fastq_cmd = f"fastqc -o www/fastqc/ -t {threads} {file}".split()
    with open(log, "a") as logf:
        run(fastq_cmd, stdout=logf, stderr=STDOUT)


def run_fastqc(
    short_reads_1,
    short_reads_2,
    threads: int,
    log: str,
):
    with open(log, "w") as logf: pass

    if os.path.exists('data/short_reads.fastq.gz'):
        fastq_cmd = f"fastqc -o www/fastqc/ -t {threads} data/short_reads.fastq.gz".split()
        with open(log, "a") as logf:
            run(fastq_cmd, stdout=logf, stderr=STDOUT)

    elif short_reads_2 != 'none': # paired end
        pool = mp.Pool(threads)
        if isinstance(short_reads_1, str):
            reads = [short_reads_1, short_reads_2]
        else:
            threads = max(len(short_reads_1 + short_reads_2) // threads, 1)
            reads = short_reads_1 + short_reads_2

        mp_results = [pool.apply_async(spawn_fastqc, args=(read, log, threads))
                    for read in
                    reads]

        for result in mp_results:
            result.get()

        pool.close()
        pool.join()
    elif short_reads_1 != 'none': # interleaved
        pool = mp.Pool(threads)
        if isinstance(short_reads_1, str):
            reads = [short_reads_1]
        else:
            threads = max(len(short_reads_1) // threads,
                        1)
            reads = short_reads_1

        mp_results = [pool.apply_async(spawn_fastqc, args=(read, log, threads))
                    for read in
                    reads]

        for result in mp_results:
            result.get()

        pool.close()
        pool.join()
    else:
        echo_cmd = 'echo "no short reads"'.split()

        with open(log, 'a') as f:
            Popen(echo_cmd, stdout=f).wait()

        with open('www/fastqc/short_reads_fastqc.html', 'w') as f:
            Popen(echo_cmd, stdout=f).wait()

    Path('www/fastqc/done').touch()


if __name__ == '__main__':
    run_fastqc(
        snakemake.config['short_reads_1'],
        snakemake.config['short_reads_2'],
        snakemake.threads,
        snakemake.log[0],
    )



from subprocess import run
import os
from pathlib import Path


def process_batch(
    batch_file_path: str,
    threads: int,
    ):
    main_directory = os.getcwd()
    with open(batch_file_path, "r") as batch_file:
        for line in batch_file:
            line = line.strip().split()
            assembly = line[0]
            identifier = line[1]
            # Assuming that batch file has reads written as interleaved files e.g. 1 2 1 2...
            # File paths must also be absolute
            forward = line[2::2]
            reverse = line[3::2]
            try:
                os.mkdir("data/" + identifier)
            except FileExistsError:
                print("Directory already exists for sample %s" % identifier)
            # Symbolically link to main aviary folder
            os.symlink(f"{main_directory}/envs", f"{main_directory}/data/{identifier}/envs")
            os.symlink(f"{main_directory}/scripts", f"{main_directory}/data/{identifier}/scripts")
            os.symlink(f"{main_directory}/.snakemake", f"{main_directory}/data/{identifier}/.snakemake")

            # Flags to specify when to change the next line
            changing_fasta = False
            changing_reads_1 = False
            changing_reads_2 = False
            changing_batch = False
            with open("%s/data/%s/template_config.yaml" % (main_directory, identifier), "w+") as new_config_file:
                with open("%s/template_config.yaml" % main_directory, 'r') as config_file:
                    for config_line in config_file:
                        if config_line.startswith("fasta"):
                            new_config_file.write(config_line)
                            changing_fasta = True
                        elif changing_fasta:
                            new_config_file.write(" " + assembly + "\n")
                            changing_fasta = False
                        elif config_line.startswith("short_reads_1"):
                            new_config_file.write(config_line)
                            changing_reads_1 = True
                        elif changing_reads_1:
                            config_line = " " + " ".join(forward) + "\n"
                            new_config_file.write(config_line)
                            changing_reads_1 = False
                        elif config_line.startswith("short_reads_2"):
                            new_config_file.write(config_line)
                            changing_reads_2 = True
                        elif changing_reads_2:
                            config_line = " " + " ".join(reverse) + "\n"
                            new_config_file.write(config_line)
                            changing_reads_2 = False
                        elif config_line.startswith("batch_file"):
                            new_config_file.write(config_line)
                            changing_batch = True
                        elif changing_batch:
                            config_line = " " + "none" + "\n"
                            new_config_file.write(config_line)
                            changing_batch = False
                        else:
                            new_config_file.write(config_line)


            os.chdir("%s/data/%s" % (main_directory, identifier))
            # Run a new snakemake process using the updated template_config.yam
            unlock_cmd = f"snakemake --unlock --use-conda --conda-prefix {main_directory}/.snakemake/ -s {main_directory}/assembly.smk recover_mags".split()
            run(unlock_cmd)

            run_cmd = f"snakemake --use-conda --conda-prefix {main_directory}/.snakemake/ -s {main_directory}/assembly.smk --cores {threads} recover_mags".split()
            run(run_cmd)
            os.chdir(main_directory)


if __name__ == "__main__":
    process_batch(snakemake.config["batch_file"], snakemake.threads)
    Path("data/done").touch()


from subprocess import Popen, PIPE, run, STDOUT
from pathlib import Path
import os
from typing import List
import gzip

def interleave(f1, f2, output_fastq:str):
    """Interleaves two (open) fastq files.
    """
    with open(output_fastq, 'ab') as output_f:
        while True:
            line = f1.readline()
            if line.strip() == b"":
                break
            output_f.write(line)
            for _ in range(3):
                output_f.write(f1.readline())
            for _ in range(4):
                output_f.write(f2.readline())

def setup_interleave(reads_1: str, reads_2: str, output_fastq: str, logf):
    logf.write(f"Interleaving reads\n")
    logf.write(f"Reads 1: {reads_1}\n")
    logf.write(f"Reads 2: {reads_2}\n")
    logf.write(f"Output: {output_fastq}\n")
    if reads_1[-2:] == "gz":
        with gzip.open(reads_1) as f1:
            with gzip.open(reads_2) as f2:
                interleave(f1, f2, output_fastq)
    else:
        with open(reads_1, 'rb') as f1:
            with open(reads_2, 'rb') as f2:
                    interleave(f1, f2, output_fastq)

def cat_reads(read_path: str, output_path: str, threads: int, log: str):
    """
    :param read_path: path to reads
    :param output_path: path to output
    :param threads: number of threads
    :param log: path to log file
    :return:
    """
    with open(log, "a") as logf:
        with open(output_path, 'a') as out:
            cat_or_zcat = 'zcat' if read_path.endswith('.gz') else 'cat'
            cat_cmd = f'{cat_or_zcat} {read_path}'.split()
            pigz_cmd = f'pigz -p {threads}'.split()

            logf.write(f"Shell style : {' '.join(cat_cmd)} | {' '.join(pigz_cmd)} > {output_path}\n")

            cat_p1 = Popen(cat_cmd, stdout=PIPE, stderr=logf)
            pigz_p2 = Popen(pigz_cmd, stdin=cat_p1.stdout, stdout=out, stderr=logf)

            pigz_p2.wait()
            cat_p1.wait()
            logf.write(f"cat return: {cat_p1.returncode}\n")
            logf.write(f"pigz return: {pigz_p2.returncode}\n")

def combine_reads(
    short_reads_1,
    short_reads_2,
    output_fastq: str,
    coassemble: bool,
    log_file: str,
    threads: int,
):
    """
    Combine reads before QC
    :param long_reads: list of long reads
    :param output_long_reads: output long reads
    :param coassemble: coassemble or not, if true we will filter all reads into the same file
    :return:
    """
    with open(log_file, 'a') as logf:
        logf.write(f"Combining reads before quality control\n")
        logf.write(f"Coassemble: {coassemble}\n")

        gzip_output = False
        if output_fastq.endswith(".gz"):
            output_fastq = output_fastq[:-3]
            gzip_output = True

        if "none" in short_reads_1 and "none" in short_reads_2:
            logf.write(f"Creating dummy files")
            Path(output_fastq).touch()


        # we've got to concatenate the files together
        if coassemble:
            if "none" not in short_reads_2 and "none" not in short_reads_1:
                for reads_1, reads_2 in zip(short_reads_1, short_reads_2):
                    if reads_1 == "none" or reads_2 == "none":
                        continue
                    
                    if not os.path.exists(reads_1):
                        logf.write(f"Short read file {reads_1} does not exist\n")
                        exit(1)
                    
                    if not os.path.exists(reads_2):
                        logf.write(f"Short read file {reads_2} does not exist\n")
                        exit(1)
                    
                    setup_interleave(reads_1, reads_2, output_fastq, logf)
            elif "none" not in short_reads_1:
                
                gzip_output = False
                if not output_fastq.endswith(".gz"):
                    output_fastq = output_fastq + '.gz'

                for reads_1 in short_reads_1:
                    if reads_1 == "none":
                        continue

                    if not os.path.exists(reads_1):
                        logf.write(f"Short read file {reads_1} does not exist\n")
                        exit(1)

                    cat_reads(reads_1, output_fastq, threads, log_file)
            elif "none" not in short_reads_2:
                gzip_output = False
                if not output_fastq.endswith(".gz"):
                    output_fastq = output_fastq + '.gz'

                for reads_2 in short_reads_2:
                    if reads_2 == "none":
                        continue

                    if not os.path.exists(reads_2):
                        logf.write(f"Short read file {reads_2} does not exist\n")
                        exit(1)

                    cat_reads(reads_2, output_fastq, threads, log_file)

        
        else:
            # if we have paired reads, we need to concatenate them together
            if "none" not in short_reads_2 and "none" not in short_reads_1:
                for reads1, reads2 in zip(short_reads_1, short_reads_2):
                    if not os.path.exists(reads1):
                        logf.write(f"Short read file {reads1} does not exist\n")
                        exit(1)

                    if not os.path.exists(reads2):
                        logf.write(f"Short read file {reads2} does not exist\n")
                        exit(1)

                    setup_interleave(reads1, reads2, output_fastq, logf)
                    break
            elif "none" not in short_reads_1:
                
                if gzip_output and short_reads_1[0].endswith(".gz"):
                    gzip_output = False
                    output_fastq = output_fastq + '.gz'

                # otherwise we just need to symlink the first file
                logf.write(f"Symlinking {short_reads_1[0]} to {output_fastq}\n")
                if not os.path.exists(short_reads_1[0]):
                    logf.write(f"Short read file {short_reads_1[0]} does not exist\n")
                    exit(1)

                os.symlink(short_reads_1[0], output_fastq)
            elif "none" not in short_reads_2:

                if gzip_output and short_reads_2[0].endswith(".gz"):
                    gzip_output = False
                    output_fastq = output_fastq + '.gz'

                # otherwise we just need to symlink the first file
                logf.write(f"Symlinking {short_reads_2[0]} to {output_fastq}\n")
                if not os.path.exists(short_reads_2[0]):
                    logf.write(f"Short read file {short_reads_2[0]} does not exist\n")
                    exit(1)

                os.symlink(short_reads_2[0], output_fastq)
            else:
                logf.write(f"Both reads_1 and reads_2 are None, Error has occured in read concatenation.\n")
                exit(1)

        if gzip_output:
            logf.write(f"Gzipping {output_fastq}\n")
            pigz_cmd = f"pigz -fp {threads} {output_fastq}".split()
            run(pigz_cmd, stderr=logf)



def run_mapping_process(
    reads_string: str, # combination of reads1 and reads2 or just reads1
    input_fasta: str,
    output_bam: str,
    output_fastq: str,
    threads: int,
    log: str,
):
    """
    :param reads_string: combination of reads1 and reads2 or just reads1
    :param input_fasta: input fasta file
    :param output_bam: output bam file
    :param output_fastq: output fastq file
    :param threads: number of threads
    :return:
    """
    with open(log, "a") as logf:
        minimap_cmd = f"minimap2 -ax sr -t {threads} {input_fasta} {reads_string}".split()
        samtools_view_cmd = f"samtools view -b -f 12 -@ {threads} -o {output_bam}".split()
        logf.write(f"Shell style : {' '.join(minimap_cmd)} | {' '.join(samtools_view_cmd)}\n")

        minimap_p1 = Popen(minimap_cmd, stdout=PIPE, stderr=logf)
        samtools_view_p2 = Popen(samtools_view_cmd, stdin=minimap_p1.stdout, stderr=logf)
        samtools_view_p2.wait()

        minimap_p1.wait()
        logf.write(f"minimap return: {minimap_p1.returncode}\n")
        logf.write(f"samtools view return: {samtools_view_p2.returncode}\n")

        # samtools index
        samtools_index_cmd = f"samtools index -@ {threads} {output_bam}".split()
        run(samtools_index_cmd, stderr=logf)

        # samtools bam2fq
        samtools_bam2fq_cmd = f"samtools bam2fq -@ {threads} -f 12 {output_bam}".split()
        pigz_cmd = f"pigz -p {threads}".split()

        logf.write(f"Shell style : {' '.join(samtools_bam2fq_cmd)} | {' '.join(pigz_cmd)} > {output_fastq}\n")
        with open(output_fastq, 'w') as output_fq:
            samtools_bam2fq_p1 = Popen(samtools_bam2fq_cmd, stdout=PIPE, stderr=logf)
            pigz_p2 = Popen(pigz_cmd, stdin=samtools_bam2fq_p1.stdout, stdout=output_fq, stderr=logf)

            samtools_bam2fq_p1.wait()
            pigz_p2.wait()
            logf.write(f"samtools bam2fq return: {samtools_bam2fq_p1.returncode}\n")
            logf.write(f"pigz return: {pigz_p2.returncode}\n")

def run_fastp(
    reads_1,
    reads_2,
    output_fastq: str,
    threads: int,
    disable_adapter_trimming: bool,
    quality_cutoff: int,
    unqualified_percent_limit: int,
    min_length: int,
    max_length: int,
    extra_fastp_params: str,
    log: str,
) -> str:
    """
    :param input_fastq: input fastq file
    :param output_fastq: output fastq file
    :param threads: number of threads
    :param disable_adapter_trimming: disable adapter trimming
    :param quality_cutoff: quality cutoff
    :param unqualified_percent_limit: unqualified percent limit
    :param min_length: minimum length
    :param max_length: maximum length
    :return:
    """
    with open(log, "a") as logf:
        if reads_1 is None and reads_2 is None:
            logf.write(f"Both reads_1 and reads_2 are None, Error has occured in read concatenation.\n")
            exit(1)

        fastp_cmd_list = ["fastp", "--stdout", "-w", str(threads), "-q", str(quality_cutoff), "-u", str(unqualified_percent_limit), "-l", str(min_length), "--length_limit", str(max_length)]
        if disable_adapter_trimming:
            fastp_cmd_list.append("--disable_adapter_trimming")

        if reads_2 is not None:
            fastp_cmd_list.extend(["-i", reads_1, "-I", reads_2])
        
        if reads_1 is not None and reads_2 is None:
            fastp_cmd_list.extend(["-i", reads_1])
        
        fastp_cmd_list.extend(extra_fastp_params.split())

        pigz_cmd = f"pigz -p {threads}".split()

        logf.write(f"Shell style : {' '.join(fastp_cmd_list)} | {' '.join(pigz_cmd)} > {output_fastq}\n")
        
        with open(output_fastq, 'w') as output_fq:
            fastp_p1 = Popen(fastp_cmd_list, stdout=PIPE, stderr=logf)
            pigz_p2 = Popen(pigz_cmd, stdin=fastp_p1.stdout, stdout=output_fq, stderr=logf)

            pigz_p2.wait()
            fastp_p1.wait()
            logf.write(f"fastp return: {fastp_p1.returncode}\n")
            logf.write(f"pigz return: {pigz_p2.returncode}\n")

    return output_fastq

def filter_illumina_reference(
    short_reads_1,
    short_reads_2,
    disable_adapter_trimming: bool,
    quality_cutoff: int,
    unqualified_percent_limit: int,
    min_length: int,
    max_length: int,
    extra_fastp_params: str,
    reference_filter: List[str],
    output_bam: str,
    output_fastq: str,
    threads: int,
    coassemble: bool,
    filtered: str,
    log: str,
    skip_qc: bool,
):
    
    if skip_qc or ("none" in short_reads_1 and "none" in short_reads_2):
        combine_reads(short_reads_1, short_reads_2, output_fastq, coassemble, log, threads)
        with open(log, 'a') as logf:
            logf.write(f"Skipping quality control\n")
        Path(filtered).touch()
        Path(output_bam).touch()
        return

    # run fastp for quality control of reads
    se1_string = None
    if "none" not in short_reads_1:
        se1_string = "data/short_reads.pre_qc.1.fastq.gz"
        combine_reads(short_reads_1, ["none"], se1_string, coassemble, log, threads)
    
    se2_string = None
    if "none" not in short_reads_2:
        se2_string = "data/short_reads.pre_qc.2.fastq.gz"
        combine_reads(["none"], short_reads_2, se2_string, coassemble, log, threads)
    
    run_fastp(
        reads_1=se1_string,
        reads_2=se2_string,
        output_fastq="data/short_reads.fastq.gz",
        threads=threads,
        disable_adapter_trimming=disable_adapter_trimming,
        quality_cutoff=quality_cutoff,
        unqualified_percent_limit=unqualified_percent_limit,
        min_length=min_length,
        max_length=max_length,
        extra_fastp_params=extra_fastp_params,
        log=log,
    )

    # Move fastp.json and fastp.html to www/
    if os.path.exists("fastp.json") and os.path.exists("fastp.html"):
        # make sure www/ exists
        if not os.path.exists("www/"):
            os.mkdir("www/")
        os.rename("fastp.html", "www/fastp.html")
        os.rename("fastp.json", "www/fastp.json")

    # remove the pre qc files
    if os.path.exists("data/short_reads.pre_qc.1.fastq.gz"):
        os.remove("data/short_reads.pre_qc.1.fastq.gz")
    if os.path.exists("data/short_reads.pre_qc.2.fastq.gz"):
        os.remove("data/short_reads.pre_qc.2.fastq.gz")

    

    reference_filter_file_string = ''
    with open(log, "a") as logf:
        if len(reference_filter) == 0:
            logf.write(f"Not performing reference filtering: {reference_filter}\n")
            Path(filtered).touch()
            Path(output_bam).touch()
            return
        
        if len(reference_filter) > 1:
            with open(f'data/reference_filter.fasta', 'w') as out:
                for reference in reference_filter:
                    # check if file exists
                    if not os.path.exists(reference):
                        logf.write(f"Reference filter file {reference} does not exist\n")
                        exit(1)

                    # concatenate accoutning for gzipped files
                    cat_or_zcat = 'zcat' if reference.endswith('.gz') else 'cat'
                    cat_cmd = f'{cat_or_zcat} {reference}'.split()

                    logf.write(f"Shell style : {' '.join(cat_cmd)} > data/reference_filter.fasta\n")

                    cat_p1 = Popen(cat_cmd, stdout=out, stderr=logf)
                    cat_p1.wait()
                    logf.write(f"cat return: {cat_p1.returncode}\n")

            # gzip the concatenated file
            pigz_cmd = f'pigz -p {threads} data/reference_filter.fasta'.split()
            logf.write(f"Shell style : {' '.join(pigz_cmd)}\n")

            pigz_p1 = Popen(pigz_cmd, stderr=logf)
            pigz_p1.wait()
            logf.write(f"pigz return: {pigz_p1.returncode}\n")

            reference_filter_file_string = f'data/reference_filter.fasta.gz'
        else:
            # make sure file exists
            if not os.path.exists(reference_filter[0]):
                logf.write(f"Reference filter file {reference_filter[0]} does not exist\n")
                exit(1)

            reference_filter_file_string = f'{reference_filter[0]}'


    run_mapping_process(
        reads_string='data/short_reads.fastq.gz',
        input_fasta=reference_filter_file_string,
        output_bam=output_bam,
        output_fastq=output_fastq,
        threads=threads,
        log=log,
    )

    # remove the reference filter file
    if os.path.exists("data/reference_filter.fasta.gz"):
        os.remove("data/reference_filter.fasta.gz")

    Path(filtered).touch()

if __name__ == '__main__':
    short_reads_1 = snakemake.input.short_reads_1
    short_reads_2 = snakemake.input.short_reads_2
    output_bam = snakemake.output.bam
    output_fastq = snakemake.output.fastq
    output_filtered = snakemake.output.filtered

    # fastp parameters
    disable_adapter_trimming = snakemake.params.disable_adapter_trimming
    quality_cutoff = snakemake.params.quality_cutoff
    unqualified_percent_limit = snakemake.params.unqualified_percent_limit
    min_length = snakemake.params.min_length
    max_length = snakemake.params.max_length
    extra_fastp_params = snakemake.params.extra_fastp_params

    coassemble = snakemake.params.coassemble
    reference_filter = snakemake.params.reference_filter
    skip_qc = snakemake.params.skip_qc

    threads = snakemake.threads
    log = snakemake.log[0]

    with open(log, "w") as logf: pass

    filter_illumina_reference(
        short_reads_1=short_reads_1,
        short_reads_2=short_reads_2,
        disable_adapter_trimming=disable_adapter_trimming,
        quality_cutoff=quality_cutoff,
        unqualified_percent_limit=unqualified_percent_limit,
        min_length=min_length,
        max_length=max_length,
        extra_fastp_params=extra_fastp_params,
        reference_filter=reference_filter,
        output_bam=output_bam,
        output_fastq=output_fastq,
        threads=threads,
        coassemble=coassemble,
        filtered=output_filtered,
        log=log,
        skip_qc=skip_qc,
    )



#!/usr/bin/env python3

import unittest
import os
import tempfile
import extern

path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
path_to_conda = os.path.join(path_to_data,'.conda')

FORWARD_READS = os.path.join(path_to_data, "wgsim.1.fq.gz")
REVERSE_READS = os.path.join(path_to_data, "wgsim.2.fq.gz")

class Tests(unittest.TestCase):
    def test_assemble_simple_inputs(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary assemble "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            self.assertTrue("fastqc" in output)
            self.assertTrue("complete_qc_short" in output)
            self.assertTrue("assemble_short_reads" in output)
            self.assertTrue("complete_assembly_with_qc" in output)
            self.assertTrue("qc_short_reads" in output)

            self.assertTrue("flye_assembly" not in output)

if __name__ == '__main__':
    unittest.main()



import pandas as pd
from subprocess import run, STDOUT
import shutil
import glob
import os
import pandas.errors as e
import datetime

def refinery():
    """
    Main function that performs the refining process iteratively passing bins between
    rosella refine and checkm until either the max number of iterations is reached
    or there are no more contaminated bins
    """
    # These will change
    current_iteration = 0
    checkm_path = snakemake.input.checkm

    # These will not change
    coverage = snakemake.input.coverage
    assembly = snakemake.input.fasta
    final_refining = snakemake.params.final_refining
    log = snakemake.log[0]

    with open(log, "w") as logf: pass

    # kmers = snakemake.input.kmers
    if os.path.isfile("data/rosella_bins/kmer_frequencies.tsv"):
        kmers = "data/rosella_bins/kmer_frequencies.tsv"
    else:
        kmers = None
    min_bin_size = snakemake.params.min_bin_size
    max_iterations = int(snakemake.params.max_iterations)
    max_retries = int(snakemake.params.max_retries)
    pplacer_threads = int(snakemake.params.pplacer_threads)
    threads = int(snakemake.threads)
    max_contamination = int(snakemake.params.max_contamination)
    contaminated_bin_folder = snakemake.params.output_folder + "/contaminated_bins"
    final_bins = snakemake.params.output_folder + "/final_bins"
    bin_folder = snakemake.params.bin_folder
    extension = snakemake.params.extension
    bin_prefix = snakemake.params.bin_prefix


    try:
        os.makedirs(contaminated_bin_folder)
    except FileExistsError:
        shutil.rmtree(contaminated_bin_folder)
        os.makedirs(contaminated_bin_folder)

    try:
        os.makedirs(final_bins)
    except FileExistsError:
        shutil.rmtree(final_bins)
        os.makedirs(final_bins)

    try:
        current_checkm = pd.read_csv(checkm_path, sep='\t', comment="[")
    except e.EmptyDataError:
        with open(log, "a") as logf:
            logf.write("No bins found in checkm file\n")
            logf.write("Skipping refinement\n")
        if max_iterations == 0:
            for bin in os.listdir(bin_folder):
                if bin.endswith(extension):
                    shutil.copy(f"{bin_folder}/{bin}", f"{final_bins}/{os.path.splitext(bin)[0]}.fna")

        open(f"{snakemake.params.output_folder}/done", "a").close()

        return

    final_checkm = current_checkm.copy().loc[current_checkm["Contamination"] <= snakemake.params.max_contamination].copy()
    final_checkm = move_finished_bins(final_checkm, bin_folder, extension, final_bins)
    previous_checkm_path = None
    unchanged_bins = None

    while current_iteration < max_iterations:
        with open(log, "a") as logf:
            # write white space for legibility
            logf.write("\n")
            logf.write("\n")

            logf.write(f"INFO: {datetime.datetime.now().strftime('%H:%M:%S')} - Refining iteration {current_iteration}\n")

        # delete previous contaminated set
        if current_iteration != 0:
            files = glob.glob(f'{contaminated_bin_folder}/*')
            [os.remove(file) for file in files]
            extension = "fna"

        # put contaminated bins in one folder. If no contaminated bins then break
        if not collect_contaminated_bins(
                current_checkm, max_contamination, bin_folder, extension, contaminated_bin_folder):
            break

        output_folder = f"{snakemake.params.output_folder}/rosella_refined_{current_iteration + 1}"

        if os.path.exists(output_folder):
            shutil.rmtree(output_folder)

        with open(log, "a") as logf:
            # write white space for legibility
            logf.write("\n")
            logf.write(f"INFO: {datetime.datetime.now().strftime('%H:%M:%S')} - Rosella refine iteration {current_iteration}\n")
        
        # Refine the contaminated bins
        kmers = refine(assembly, coverage, kmers, checkm_path,
                   contaminated_bin_folder, extension, min_bin_size,
                   threads, output_folder, max_contamination, max_retries, f"{bin_prefix}_refined_{current_iteration + 1}", log)

        # update the bin folder variable to the current refined folder
        bin_folder = f"{output_folder}/refined_bins/"
        unchanged_bins = f"{output_folder}/unchanged_bins/"

        # move the unchanged bins to the final bins folder
        if os.path.exists(unchanged_bins):
            for mag in os.listdir(unchanged_bins):
                shutil.move(f"{unchanged_bins}/{mag}", f"{final_bins}/{os.path.splitext(mag)[0]}.fna")

        # retrieve the checkm results for the refined bins
        try:
            with open(log, "a") as logf:
                # write white space for legibility
                logf.write("\n")
                logf.write(f"INFO: {datetime.datetime.now().strftime('%H:%M:%S')} - CheckM iteration {current_iteration}\n")
            # count how many bins in bin_folder, bins end in 'fna'
            bin_count = 0
            for bin_file in os.listdir(bin_folder):
                if bin_file.endswith("fna"):
                    bin_count += 1
            
            if bin_count == 0:
                with open(log, "a") as logf:
                    logf.write("No bins to refine\n")
                    logf.write("Skipping refinement\n")
                # No bins to refine, break out and move on
                break
            else:
                with open(log, "a") as logf:
                    logf.write(f"Running CheckM on {bin_count} bins\n")

            get_checkm_results(bin_folder, threads, pplacer_threads, log, final_refining)
            # update the checkm results and counter
            checkm_path = f"{bin_folder}/checkm.out"
            current_checkm = pd.read_csv(checkm_path, sep='\t', comment='[')

            bins_to_keep = current_checkm.copy().loc[current_checkm["Contamination"] <= max_contamination]
            bins_to_keep = move_finished_bins(bins_to_keep, bin_folder, "fna", final_bins, None) # None because bin tag handles refined bin name
            final_checkm = pd.concat([final_checkm, bins_to_keep])

            # also handle unchanged bins
            if os.path.exists(unchanged_bins):
                # list the bins in unchanged_bins
                if previous_checkm_path is None:
                    previous_checkm_path = snakemake.input.checkm
                
                bin_id_accessor_str = "Bin Id"

                previous_checkm = pd.read_csv(previous_checkm_path, sep='\t', comment='[')
                if "Bin Id" not in previous_checkm.columns:
                    # checkm2 input
                    bin_id_accessor_str = "Name"

                bin_ids = []
                for bin_path in os.listdir(unchanged_bins):
                    # copy the bin to the final bins folder
                    bin_id = os.path.splitext(bin_path)[0]
                    bin_ids.append(bin_id)
                    shutil.copy(f"{unchanged_bins}/{bin}", f"{final_bins}/{bin_id}.fna")
                    # add the bin to the final checkm
                    # row = previous_checkm.copy().loc[previous_checkm["Bin Id"] == bin_id]
                    # final_checkm = pd.concat([final_checkm, row])
                final_checkm = pd.concat([final_checkm, previous_checkm.copy().loc[previous_checkm[bin_id_accessor_str].isin(bin_ids)]])

            previous_checkm_path = checkm_path

            with open(log, "a") as logf:
                logf.write(f"INFO: {datetime.datetime.now().strftime('%H:%M:%S')} - Ending iteration: {current_iteration}\n")

            current_iteration += 1
        except FileNotFoundError:
            with open(log, "a") as logf:
                logf.write("No bins to refine\n")
                logf.write(f"Ending refine on iteration: {current_iteration}\n")
            # No bins to refine, break out and move on
            break
    

    if final_refining:
        final_checkm.to_csv("data/checkm.out", sep='\t', index=False)
        final_output_folder = "bins/final_bins"
        if os.path.exists("bins/"): # remove pre-existing output, preventing bin duplication
            shutil.rmtree("bins/")
        os.makedirs("bins/")
        if not os.path.exists(final_output_folder):
            os.symlink("../" + final_bins, final_output_folder, target_is_directory=True)
        elif not os.path.islink(final_output_folder):
            os.rmdir(final_output_folder)
            os.symlink("../" + final_bins, final_output_folder, target_is_directory=True)
        final_checkm.to_csv("bins/checkm.out", sep='\t', index=False)
    else:
        with open(log, "a") as logf:
            logf.write("Refinery finished.\n")
        
        # final_checkm.to_csv(f"{snakemake.params.output_folder}/intermediate_checkm.out", sep='\t', index=False)
        open(f"{snakemake.params.output_folder}/done", "a").close()
    
    # in output folder, we then delete directories starting with rosella_refined_
    for folder in os.listdir(snakemake.params.output_folder):
        if folder.startswith("rosella_refined_"):
            shutil.rmtree(f"{snakemake.params.output_folder}/{folder}")



def move_finished_bins(
        checkm,
        input_folder,
        input_extension,
        output_folder,
        current_iteration = None
):
    try:
        bins = checkm["Bin Id"]
        column_name = "Bin Id"
    except KeyError:
        # checkm2 input
        bins = checkm["Name"]
        column_name = "Name"

    new_bin_ids = []
    for bin_id in bins:
        # print(f"{input_folder}/{bin_id}")
        if current_iteration is not None:
            new_bin_id = f"{bin_id}_{current_iteration}"
        else:
            new_bin_id = bin_id

        new_bin_ids.append(new_bin_id)
        try:
            shutil.copy(f"{input_folder}/{bin_id}.{input_extension}", f"{output_folder}/{new_bin_id}.fna")
        except FileNotFoundError:
            # File already moved
            continue

    checkm[column_name] = new_bin_ids

    return checkm


def collect_contaminated_bins(
        checkm,
        max_contamination,
        bin_folder,
        extension,
        output_folder
):
    contaminated = checkm[checkm["Contamination"] > max_contamination]

    # check that there is at least one contaminated bin
    if contaminated.shape[0] == 0: return False

    try:
        bins = contaminated["Bin Id"]
    except KeyError:
        # checkm2 input
        bins = contaminated["Name"]


    for bin_id in bins:
        try:
            shutil.copy(f"{bin_folder}/{bin_id}.{extension}", f"{output_folder}/{bin_id}.{extension}")
        except FileNotFoundError:
            # file already moved
            continue

    return True


def refine(
    assembly,
    coverage,
    kmers,
    checkm,
    bin_folder,
    extension,
    min_bin_size,
    threads,
    output_folder,
    max_contamination,
    max_retries,
    bin_tag,
    log,
):
    if kmers is None:
        rosella_cmd = f"rosella refine -r {assembly} -C {coverage} -d {bin_folder} -x {extension} --checkm-results {checkm} --max-contamination {max_contamination} --max-retries {max_retries} --min-bin-size {min_bin_size} -t {threads} -o {output_folder} --bin-tag {bin_tag}".split()
        with open(log, "a") as logf:
            run(rosella_cmd, stdout=logf, stderr=STDOUT)

        os.makedirs("data/rosella_bins/", exist_ok=True)
        shutil.copyfile(f"{output_folder}/kmer_frequencies.tsv", f"data/rosella_bins/kmer_frequencies.tsv")
        kmers = "data/rosella_bins/kmer_frequencies.tsv"
    else:
        rosella_cmd = f"rosella refine -r {assembly} -C {coverage} --kmer-frequency-file {kmers} -d {bin_folder} -x {extension} --checkm-results {checkm} --max-contamination {max_contamination} --max-retries {max_retries} --min-bin-size {min_bin_size} -t {threads} -o {output_folder} --bin-tag {bin_tag}".split()

        with open(log, "a") as logf:
            run(rosella_cmd, stdout=logf, stderr=STDOUT)

    return kmers

def get_checkm_results(
        refined_folder,
        threads,
        pplacer_threads,
        log,
        final_refining=False,
):

    checkm_cmd = f"checkm lineage_wf -t {threads} --pplacer_threads {pplacer_threads} -x fna --tab_table -f {refined_folder}/checkm.out {refined_folder} {refined_folder}/checkm".split()
    with open(log, "a") as logf:
        run(checkm_cmd, stdout=logf, stderr=STDOUT)

    if final_refining:
        checkm_qa_cmd = f"checkm qa -o 2 --tab_table -f {refined_folder}/checkm.out {refined_folder}/checkm/lineage.ms {refined_folder}/checkm/".split()
        with open(log, "a") as logf:
            run(checkm_qa_cmd, stdout=logf, stderr=STDOUT)


if __name__ == '__main__':
    refinery()


from subprocess import Popen, PIPE, run
import os
import logging


def run_mapping_process(
    reads_string: str,
    input_fasta: str,
    output_bam: str,
    threads: int,
    log: str,
):
    """
    :param reads_string: combination of reads1 and reads2 or just reads1
    :param input_fasta: input fasta file
    :param output_bam: output bam file
    :param output_fastq: output fastq file
    :param threads: number of threads
    :return:
    """
    with open(log, "a") as logf:
        minimap_cmd = f"minimap2 -ax sr -t {threads} {input_fasta} {reads_string}".split()
        samtools_view_cmd = f"samtools view -b -F 4 -@ {threads}".split()
        samtools_sort_cmd = f"samtools sort -@ {threads} -o {output_bam} -".split()
        logf.write(f"Shell style : {' '.join(minimap_cmd)} | {' '.join(samtools_view_cmd)} | {' '.join(samtools_sort_cmd)}\n")

        minimap_p1 = Popen(minimap_cmd, stdout=PIPE, stderr=logf) # stderr=PIPE optional, dd is chatty
        samtools_view_p2 = Popen(samtools_view_cmd, stdin=minimap_p1.stdout, stdout=PIPE, stderr=logf)
        samtools_sort_p3 = Popen(samtools_sort_cmd, stdin=samtools_view_p2.stdout, stderr=logf)
        samtools_sort_p3.wait()

        # thoretically p1 and p2 may still be running, this ensures we are collecting their return codes
        minimap_p1.wait()
        samtools_view_p2.wait()
        logf.write(f"minimap return: {minimap_p1.returncode}\n")
        logf.write(f"samtools view return: {samtools_view_p2.returncode}\n")
        logf.write(f"samtools sort return: {samtools_sort_p3.returncode}\n")

        # samtools index
        samtools_index_cmd = f"samtools index -@ {threads} {output_bam}".split()
        run(samtools_index_cmd, stderr=logf)

def generate_pilon_sort(
    short_reads_1,
    short_reads_2,
    input_fasta: str,
    output_bam: str,
    threads: int,
    coassemble: bool,
    log: str,
):
    logging.info("Generating BAM files for pilon...")
    if os.path.exists('data/short_reads.fastq.gz'):
        run_mapping_process(
            reads_string='data/short_reads.fastq.gz',
            input_fasta=input_fasta,
            output_bam=output_bam,
            threads=threads,
            log=log,
        )

    elif short_reads_2 != 'none':
        if len(short_reads_2) == 1 or not coassemble:
            pe1 = short_reads_1[0]
            pe2 = short_reads_2[0]
        else:
            if not os.path.exists("data/short_reads.1.fastq.gz"):
                with open(log, "a") as logf:
                    for reads1, reads2 in zip(short_reads_1, short_reads_2):
                        with open("data/short_reads.1.fastq.gz", "a") as f:
                            run(f"cat {reads1}", stdout=f, stderr=logf)

                        with open("data/short_reads.2.fastq.gz", "a") as f:
                            run(f"cat {reads2}", stdout=f, stderr=logf)
            pe1 = "data/short_reads.1.fastq.gz"
            pe2 = "data/short_reads.2.fastq.gz"

        reads_string = f"{pe1} {pe2}"
        run_mapping_process(
            reads_string=reads_string,
            input_fasta=input_fasta,
            output_bam=output_bam,
            threads=threads,
            log=log,
        )

        if os.path.exists("data/short_reads.1.fastq.gz"):
            os.remove("data/short_reads.1.fastq.gz")
            os.remove("data/short_reads.2.fastq.gz")

    elif short_reads_1 != 'none':
        if len(short_reads_1) == 1:
            pe1 = short_reads_1[0]
        else:
            if not os.path.exists("data/short_reads.1.fastq.gz") or not coassemble:
                for reads1 in short_reads_1:
                    with open(log, "a") as logf:
                        with open("data/short_reads.1.fastq.gz", "a") as f:
                            run(f"cat {reads1}", stdout=f, stderr=logf)
            pe1 = "data/short_reads.1.fastq.gz"


        run_mapping_process(
            reads_string=pe1,
            input_fasta=input_fasta,
            output_bam=output_bam,
            threads=threads,
            log=log,
        )

        if os.path.exists("data/short_reads.1.fastq.gz"):
            os.remove("data/short_reads.1.fastq.gz")


if __name__ == "__main__":
    short_reads_1 = snakemake.config['short_reads_1']
    short_reads_2 = snakemake.config['short_reads_2']
    input_fasta = snakemake.input.fasta
    output_bam = snakemake.output.bam
    threads = snakemake.threads
    coassemble = snakemake.params.coassemble
    log = snakemake.log[0]

    with open(log, "w") as logf: pass

    generate_pilon_sort(
        short_reads_1,
        short_reads_2,
        input_fasta,
        output_bam,
        threads,
        coassemble,
        log,
    )



#!/usr/bin/env python3

import unittest
import os
import tempfile
import extern
from snakemake import load_configfile

path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
path_to_conda = os.path.join(path_to_data,'.conda')

FORWARD_READS = os.path.join(path_to_data, "wgsim.1.fq.gz")
REVERSE_READS = os.path.join(path_to_data, "wgsim.2.fq.gz")
ASSEMBLY = os.path.join(path_to_data, "assembly.fasta")

class Tests(unittest.TestCase):
    def test_recover_simple_inputs(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            # Binners
            self.assertTrue("prepare_binning_files" in output)
            self.assertTrue("get_bam_indices" in output)
            self.assertTrue("metabat_sens" in output)
            self.assertTrue("metabat_spec" in output)
            self.assertTrue("metabat_ssens" in output)
            self.assertTrue("metabat_sspec" in output)
            self.assertTrue("metabat2" in output)
            self.assertTrue("maxbin2" not in output)
            self.assertTrue("rosella" in output)
            self.assertTrue("semibin" in output)
            self.assertTrue("vamb" in output)
            self.assertTrue("concoct" not in output)
            self.assertTrue("das_tool" in output)

            # Refinery
            self.assertTrue("checkm_metabat2" in output)
            self.assertTrue("refine_metabat2" in output)
            self.assertTrue("checkm_rosella" in output)
            self.assertTrue("refine_rosella" in output)
            self.assertTrue("checkm_semibin" in output)
            self.assertTrue("refine_semibin" in output)
            self.assertTrue("checkm_das_tool" in output)
            self.assertTrue("refine_dastool" in output)

            # Extras
            self.assertTrue("checkm2" in output)
            self.assertTrue("gtdbtk" in output)
            self.assertTrue("get_abundances" in output)
            self.assertTrue("singlem_pipe_reads" in output)
            self.assertTrue("singlem_appraise" in output)
            self.assertTrue("finalise_stats" in output)
            self.assertTrue("recover_mags" in output)

            # Unnecessary
            self.assertTrue("complete_assembly_with_qc" not in output)

    def test_recover_skip_binners(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--skip-binners metabat "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            # Binners
            self.assertTrue("prepare_binning_files" in output)
            self.assertTrue("get_bam_indices" in output)
            self.assertTrue("metabat_sens" not in output)
            self.assertTrue("metabat_spec" not in output)
            self.assertTrue("metabat_ssens" not in output)
            self.assertTrue("metabat_sspec" not in output)
            self.assertTrue("metabat2" not in output)
            self.assertTrue("maxbin2" not in output)
            self.assertTrue("rosella" in output)
            self.assertTrue("semibin" in output)
            self.assertTrue("vamb" in output)
            self.assertTrue("concoct" not in output)
            self.assertTrue("das_tool" in output)

            # Refinery
            self.assertTrue("checkm_metabat2" not in output)
            self.assertTrue("refine_metabat2" not in output)
            self.assertTrue("checkm_rosella" in output)
            self.assertTrue("refine_rosella" in output)
            self.assertTrue("checkm_semibin" in output)
            self.assertTrue("refine_semibin" in output)
            self.assertTrue("checkm_das_tool" in output)
            self.assertTrue("refine_dastool" in output)

            # Extras
            self.assertTrue("checkm2" in output)
            self.assertTrue("gtdbtk" in output)
            self.assertTrue("get_abundances" in output)
            self.assertTrue("singlem_pipe_reads" in output)
            self.assertTrue("singlem_appraise" in output)
            self.assertTrue("finalise_stats" in output)
            self.assertTrue("recover_mags" in output)

            # Unnecessary
            self.assertTrue("complete_assembly_with_qc" not in output)

    def test_recover_no_singlem(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--skip-singlem "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            # Binners
            self.assertTrue("prepare_binning_files" in output)
            self.assertTrue("get_bam_indices" in output)
            self.assertTrue("metabat_sens" in output)
            self.assertTrue("metabat_spec" in output)
            self.assertTrue("metabat_ssens" in output)
            self.assertTrue("metabat_sspec" in output)
            self.assertTrue("metabat2" in output)
            self.assertTrue("maxbin2" not in output)
            self.assertTrue("rosella" in output)
            self.assertTrue("semibin" in output)
            self.assertTrue("vamb" in output)
            self.assertTrue("concoct" not in output)
            self.assertTrue("das_tool" in output)

            # Refinery
            self.assertTrue("checkm_metabat2" in output)
            self.assertTrue("refine_metabat2" in output)
            self.assertTrue("checkm_rosella" in output)
            self.assertTrue("refine_rosella" in output)
            self.assertTrue("checkm_semibin" in output)
            self.assertTrue("refine_semibin" in output)
            self.assertTrue("checkm_das_tool" in output)
            self.assertTrue("refine_dastool" in output)

            # Extras
            self.assertTrue("checkm2" in output)
            self.assertTrue("gtdbtk" in output)
            self.assertTrue("get_abundances" in output)
            self.assertTrue("singlem_pipe_reads" not in output)
            self.assertTrue("singlem_appraise" not in output)
            self.assertTrue("finalise_stats" in output)
            self.assertTrue("recover_mags" in output)

            # Unnecessary
            self.assertTrue("complete_assembly_with_qc" not in output)

    def test_recover_no_abundances(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--skip-abundances "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            # Binners
            self.assertTrue("prepare_binning_files" in output)
            self.assertTrue("get_bam_indices" in output)
            self.assertTrue("metabat_sens" in output)
            self.assertTrue("metabat_spec" in output)
            self.assertTrue("metabat_ssens" in output)
            self.assertTrue("metabat_sspec" in output)
            self.assertTrue("metabat2" in output)
            self.assertTrue("maxbin2" not in output)
            self.assertTrue("rosella" in output)
            self.assertTrue("semibin" in output)
            self.assertTrue("vamb" in output)
            self.assertTrue("concoct" not in output)
            self.assertTrue("das_tool" in output)

            # Refinery
            self.assertTrue("checkm_metabat2" in output)
            self.assertTrue("refine_metabat2" in output)
            self.assertTrue("checkm_rosella" in output)
            self.assertTrue("refine_rosella" in output)
            self.assertTrue("checkm_semibin" in output)
            self.assertTrue("refine_semibin" in output)
            self.assertTrue("checkm_das_tool" in output)
            self.assertTrue("refine_dastool" in output)

            # Extras
            self.assertTrue("checkm2" in output)
            self.assertTrue("gtdbtk" in output)
            self.assertTrue("get_abundances" not in output)
            self.assertTrue("singlem_pipe_reads" in output)
            self.assertTrue("singlem_appraise" in output)
            self.assertTrue("finalise_stats" in output)
            self.assertTrue("recover_mags" in output)

            # Unnecessary
            self.assertTrue("complete_assembly_with_qc" not in output)

    def test_recover_no_taxonomy(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--skip-taxonomy "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            # Binners
            self.assertTrue("prepare_binning_files" in output)
            self.assertTrue("get_bam_indices" in output)
            self.assertTrue("metabat_sens" in output)
            self.assertTrue("metabat_spec" in output)
            self.assertTrue("metabat_ssens" in output)
            self.assertTrue("metabat_sspec" in output)
            self.assertTrue("metabat2" in output)
            self.assertTrue("maxbin2" not in output)
            self.assertTrue("rosella" in output)
            self.assertTrue("semibin" in output)
            self.assertTrue("vamb" in output)
            self.assertTrue("concoct" not in output)
            self.assertTrue("das_tool" in output)

            # Refinery
            self.assertTrue("checkm_metabat2" in output)
            self.assertTrue("refine_metabat2" in output)
            self.assertTrue("checkm_rosella" in output)
            self.assertTrue("refine_rosella" in output)
            self.assertTrue("checkm_semibin" in output)
            self.assertTrue("refine_semibin" in output)
            self.assertTrue("checkm_das_tool" in output)
            self.assertTrue("refine_dastool" in output)

            # Extras
            self.assertTrue("checkm2" in output)
            self.assertTrue("gtdbtk" not in output)
            self.assertTrue("get_abundances" in output)
            self.assertTrue("singlem_pipe_reads" in output)
            self.assertTrue("singlem_appraise" in output)
            self.assertTrue("finalise_stats" in output)
            self.assertTrue("recover_mags" in output)

            # Unnecessary
            self.assertTrue("complete_assembly_with_qc" not in output)

    def test_recover_binning_only(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--binning-only "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun --tmpdir {tmpdir} "
                f"--snakemake-cmds \" --quiet\" "
            )
            output = extern.run(cmd)

            # Binners
            self.assertTrue("prepare_binning_files" in output)
            self.assertTrue("get_bam_indices" in output)
            self.assertTrue("metabat_sens" in output)
            self.assertTrue("metabat_spec" in output)
            self.assertTrue("metabat_ssens" in output)
            self.assertTrue("metabat_sspec" in output)
            self.assertTrue("metabat2" in output)
            self.assertTrue("maxbin2" not in output)
            self.assertTrue("rosella" in output)
            self.assertTrue("semibin" in output)
            self.assertTrue("vamb" in output)
            self.assertTrue("concoct" not in output)
            self.assertTrue("das_tool" in output)

            # Refinery
            self.assertTrue("checkm_metabat2" in output)
            self.assertTrue("refine_metabat2" in output)
            self.assertTrue("checkm_rosella" in output)
            self.assertTrue("refine_rosella" in output)
            self.assertTrue("checkm_semibin" in output)
            self.assertTrue("refine_semibin" in output)
            self.assertTrue("checkm_das_tool" in output)
            self.assertTrue("refine_dastool" in output)

            # Extras
            self.assertTrue("checkm2" in output)
            self.assertTrue("gtdbtk" not in output)
            self.assertTrue("get_abundances" not in output)
            self.assertTrue("singlem_pipe_reads" not in output)
            self.assertTrue("singlem_appraise" not in output)
            self.assertTrue("finalise_stats" in output)
            self.assertTrue("recover_mags" in output)

            # Unnecessary
            self.assertTrue("complete_assembly_with_qc" not in output)

    def test_recover_config(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd = (
                f"GTDBTK_DATA_PATH=. "
                f"CHECKM2DB=. "
                f"EGGNOG_DATA_DIR=. "
                f"SINGLEM_METAPACKAGE_PATH=. "
                f"aviary recover "
                f"--refinery-max-iterations 3 "
                f"--assembly {ASSEMBLY} "
                f"-1 {FORWARD_READS} "
                f"-2 {REVERSE_READS} "
                f"--output {tmpdir}/test --tmpdir {tmpdir} "
                f"--conda-prefix {path_to_conda} "
                f"--dryrun "
                f"--snakemake-cmds \" --quiet\" "
            )
            extern.run(cmd)

            config_path = os.path.join(tmpdir, "test", "config.yaml")
            self.assertTrue(os.path.exists(config_path))
            config = load_configfile(config_path)

            self.assertEqual(config["refinery_max_iterations"], 3)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

#=======================================================================
# Author:
#
# Unit tests.
#
# Copyright
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License.
# If not, see <http://www.gnu.org/licenses/>.
#=======================================================================

import unittest
import os.path
import subprocess
import shutil

data = os.path.join(os.path.dirname(__file__), 'data')
path_to_conda = os.path.join(data,'.conda')

class Tests(unittest.TestCase):
    def setup_output_dir(self, output_dir):
        try:
            shutil.rmtree(output_dir)
        except FileNotFoundError:
            pass
        os.makedirs(output_dir)

    def test_short_read_assembly(self):
        output_dir = os.path.join("example", "test_short_read_assembly")
        self.setup_output_dir(output_dir)
        cmd = (
            f"aviary assemble "
            f"-o {output_dir}/aviary_out "
            f"-1 {data}/wgsim.1.fq.gz "
            f"-2 {data}/wgsim.2.fq.gz "
            f"--conda-prefix {path_to_conda} "
            f"-n 32 -t 32 "
        )
        subprocess.run(cmd, shell=True, check=True)

        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out"))
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/data/final_contigs.fasta"))
        self.assertTrue(os.path.islink(f"{output_dir}/aviary_out/assembly/final_contigs.fasta"))

    def test_long_read_assembly(self):
        output_dir = os.path.join("example", "test_long_read_assembly")
        self.setup_output_dir(output_dir)
        cmd = (
            f"aviary assemble "
            f"-o {output_dir}/aviary_out "
            f"-1 {data}/wgsim.1.fq.gz "
            f"-2 {data}/wgsim.2.fq.gz "
            f"-l {data}/pbsim.fq.gz "
            f"--longread-type ont "
            f"--min-read-size 10 --min-mean-q 1 "
            f"--conda-prefix {path_to_conda} "
            f"-n 32 -t 32 "
        )
        subprocess.run(cmd, shell=True, check=True)

        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out"))
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/data/final_contigs.fasta"))
        self.assertTrue(os.path.islink(f"{output_dir}/aviary_out/assembly/final_contigs.fasta"))

    def test_short_read_recovery(self):
        output_dir = os.path.join("example", "test_short_read_recovery")
        self.setup_output_dir(output_dir)
        cmd = (
            f"aviary recover "
            f"-o {output_dir}/aviary_out "
            f"-1 {data}/wgsim.1.fq.gz "
            f"-2 {data}/wgsim.2.fq.gz "
            f"--conda-prefix {path_to_conda} "
            f"-n 32 -t 32 "
        )
        subprocess.run(cmd, shell=True, check=True)

        bin_info_path = f"{output_dir}/aviary_out/bins/bin_info.tsv"
        self.assertTrue(os.path.isfile(bin_info_path))
        with open(bin_info_path) as f:
            num_lines = sum(1 for _ in f)
        self.assertTrue(num_lines > 1)

        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/data/final_contigs.fasta"))
        self.assertTrue(os.path.islink(f"{output_dir}/aviary_out/assembly/final_contigs.fasta"))

        self.assertTrue(os.path.islink(f"{output_dir}/aviary_out/diversity"))
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/diversity/metagenome.combined_otu_table.csv"))
        self.assertTrue(os.path.getsize(f"{output_dir}/aviary_out/diversity/metagenome.combined_otu_table.csv") > 0)
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/diversity/singlem_appraisal.tsv"))
        self.assertTrue(os.path.getsize(f"{output_dir}/aviary_out/diversity/singlem_appraisal.tsv") > 0)
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/diversity/singlem_appraise.svg"))

    def test_long_read_recovery(self):
        output_dir = os.path.join("example", "test_long_read_recovery")
        self.setup_output_dir(output_dir)
        cmd = (
            f"aviary recover "
            f"-o {output_dir}/aviary_out "
            f"-1 {data}/wgsim.1.fq.gz "
            f"-2 {data}/wgsim.2.fq.gz "
            f"-l {data}/pbsim.fq.gz "
            f"--longread-type ont "
            f"--min-read-size 10 --min-mean-q 1 "
            f"--conda-prefix {path_to_conda} "
            f"-n 32 -t 32 "
        )
        subprocess.run(cmd, shell=True, check=True)

        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out"))
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/data/final_contigs.fasta"))
        self.assertTrue(os.path.islink(f"{output_dir}/aviary_out/assembly/final_contigs.fasta"))

        self.assertTrue(os.path.islink(f"{output_dir}/aviary_out/diversity"))
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/diversity/metagenome.combined_otu_table.csv"))
        self.assertTrue(os.path.getsize(f"{output_dir}/aviary_out/diversity/metagenome.combined_otu_table.csv") > 0)
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/diversity/singlem_appraisal.tsv"))
        self.assertTrue(os.path.getsize(f"{output_dir}/aviary_out/diversity/singlem_appraisal.tsv") > 0)
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/diversity/singlem_appraise.svg"))

    def test_short_read_recovery_fast(self):
        output_dir = os.path.join("example", "test_short_read_recovery_fast")
        self.setup_output_dir(output_dir)
        cmd = (
            f"aviary recover "
            f"--assembly {data}/assembly.fasta "
            f"-o {output_dir}/aviary_out "
            f"-1 {data}/wgsim.1.fq.gz "
            f"-2 {data}/wgsim.2.fq.gz "
            f"--binning-only "
            f"--skip-binners rosella vamb metabat "
            f"--skip-qc "
            f"--refinery-max-iterations 0 "
            f"--conda-prefix {path_to_conda} "
            f"-n 32 -t 32 "
        )
        subprocess.run(cmd, shell=True, check=True)

        bin_info_path = f"{output_dir}/aviary_out/bins/bin_info.tsv"
        self.assertTrue(os.path.isfile(bin_info_path))
        with open(bin_info_path) as f:
            num_lines = sum(1 for _ in f)
        self.assertEqual(num_lines, 3)

        self.assertFalse(os.path.isfile(f"{output_dir}/aviary_out/data/final_contigs.fasta"))

    def test_short_read_recovery_queue_submission(self):
        output_dir = os.path.join("example", "test_short_read_recovery_queue_submission")
        self.setup_output_dir(output_dir)

        cmd = (
            f"aviary recover "
            f"-o {output_dir}/aviary_out "
            f"-1 {data}/wgsim.1.fq.gz "
            f"-2 {data}/wgsim.2.fq.gz "
            f"--conda-prefix {path_to_conda} "
            f"-n 32 -t 32 "
            f"--snakemake-profile mqsub --cluster-retries 3 "
        )
        subprocess.run(cmd, shell=True, check=True)

        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/data/final_contigs.fasta"))

        bin_info_path = f"{output_dir}/aviary_out/bins/bin_info.tsv"
        self.assertTrue(os.path.isfile(bin_info_path))
        with open(bin_info_path) as f:
            num_lines = sum(1 for _ in f)
        self.assertEqual(num_lines, 3)

    def test_batch_recovery(self):
        output_dir = os.path.join("example", "test_batch_recovery")
        self.setup_output_dir(output_dir)
        cmd = (
            f"aviary batch "
            f"-o {output_dir}/aviary_out "
            f"-f {data}/example_batch.tsv "
            f"--conda-prefix {path_to_conda} "
            f"--skip-binners rosella vamb metabat "
            f"--skip-qc "
            f"--refinery-max-iterations 0 "
            f"--min-read-size 10 --min-mean-q 1 "
            f"-n 32 -t 32 "
        )
        subprocess.run(cmd, shell=True, check=True)

        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/sample_1/data/final_contigs.fasta"))
        self.assertTrue(os.path.isfile(f"{output_dir}/aviary_out/sample_2/data/final_contigs.fasta"))

        bin_info_path_1 = f"{output_dir}/aviary_out/sample_1/bins/bin_info.tsv"
        bin_info_path_2 = f"{output_dir}/aviary_out/sample_2/bins/bin_info.tsv"
        self.assertTrue(os.path.isfile(bin_info_path_1))
        self.assertTrue(os.path.isfile(bin_info_path_2))
        with open(bin_info_path_1) as f:
            num_lines = sum(1 for _ in f)
        self.assertEqual(num_lines, 3)

        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out/aviary_cluster_ani_0.95"))
        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out/aviary_cluster_ani_0.97"))
        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out/aviary_cluster_ani_0.99"))

        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out/aviary_cluster_ani_0.95/pangenomes"))
        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out/aviary_cluster_ani_0.97/pangenomes"))
        self.assertTrue(os.path.isdir(f"{output_dir}/aviary_out/aviary_cluster_ani_0.99/pangenomes"))


if __name__ == "__main__":
    unittest.main()
